# -*- coding: utf-8 -*-
# ui_tk.py
#
# Copyright © 2009-2010 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.


from pyglossary.core import homeDir
from pyglossary.glossary import *
from pyglossary.text_utils import urlToPath
from .base import *
from os.path import join
import logging
import traceback

from typing import Union, Dict

import tkinter as tk
from tkinter import filedialog
from tkinter import tix
from tkinter import ttk
from tkinter import font as tkFont

log = logging.getLogger("root")

# startBold = "\x1b[1m"  # Start Bold #len=4
# startUnderline = "\x1b[4m"  # Start Underline #len=4
endFormat = "\x1b[0;0;0m"  # End Format #len=8
# redOnGray = "\x1b[0;1;31;47m"
startRed = "\x1b[31m"


bitmapLogo = join(dataDir, "res", "pyglossary.ico") if "nt" == os.name \
	else "@" + join(dataDir, "res", "pyglossary.xbm")


def set_window_icon(window):
	# window.wm_iconbitmap(bitmap=bitmapLogo)
	window.iconphoto(
		True,
		tk.PhotoImage(file=join(dataDir, "res", "pyglossary.png")),
	)

def decodeGeometry(gs):
	"""
		example for gs: "253x252+30+684"
		returns (x, y, w, h)
	"""
	p = gs.split("+")
	w, h = p[0].split("x")
	return (int(p[1]), int(p[2]), int(w), int(h))

def encodeGeometry(x, y, w, h):
	return "%sx%s+%s+%s" % (w, h, x, y)

def encodeLocation(x, y):
	return "+%s+%s" % (x, y)


def centerWindow(win):
	"""
	centers a tkinter window
	:param win: the root or Toplevel window to center
	"""
	win.update_idletasks()
	width = win.winfo_width()
	frm_width = win.winfo_rootx() - win.winfo_x()
	win_width = width + 2 * frm_width
	height = win.winfo_height()
	titlebar_height = win.winfo_rooty() - win.winfo_y()
	win_height = height + titlebar_height + frm_width
	x = win.winfo_screenwidth() // 2 - win_width // 2
	y = win.winfo_screenheight() // 2 - win_height // 2
	win.geometry(encodeGeometry(x, y, width, height))
	win.deiconify()


class TkTextLogHandler(logging.Handler):
	def __init__(self, tktext):
		logging.Handler.__init__(self)
		#####
		tktext.tag_config("CRITICAL", foreground="#ff0000")
		tktext.tag_config("ERROR", foreground="#ff0000")
		tktext.tag_config("WARNING", foreground="#ffff00")
		tktext.tag_config("INFO", foreground="#00ff00")
		tktext.tag_config("DEBUG", foreground="#ffffff")
		###
		self.tktext = tktext

	def emit(self, record):
		msg = record.getMessage()
		###
		if record.exc_info:
			_type, value, tback = record.exc_info
			tback_text = "".join(
				traceback.format_exception(_type, value, tback)
			)
			if msg:
				msg += "\n"
			msg += tback_text
		###
		self.tktext.insert(
			"end",
			msg + "\n",
			record.levelname,
		)


# Monkey-patch Tkinter
# http://stackoverflow.com/questions/5191830/python-exception-logging
def CallWrapper__call__(self, *args):
	"""
		Apply first function SUBST to arguments, than FUNC.
	"""
	if self.subst:
		args = self.subst(*args)
	try:
		return self.func(*args)
	except:
		log.exception("Exception in Tkinter callback:")
tk.CallWrapper.__call__ = CallWrapper__call__


class ProgressBar(tix.Frame):
	"""
	This comes from John Grayson's book "Python and Tkinter programming"
	Edited by Saeed Rasooli
	"""
	def __init__(
		self,
		rootWin=None,
		orientation="horizontal",
		min_=0,
		max_=100,
		width=100,
		height=18,
		appearance="sunken",
		fillColor="blue",
		background="gray",
		labelColor="yellow",
		labelFont="Verdana",
		labelFormat="%d%%",
		value=0,
		bd=2,
	):
		# preserve various values
		self.rootWin = rootWin
		self.orientation = orientation
		self.min = min_
		self.max = max_
		self.width = width
		self.height = height
		self.fillColor = fillColor
		self.labelFont = labelFont
		self.labelColor = labelColor
		self.background = background
		self.labelFormat = labelFormat
		self.value = value
		tix.Frame.__init__(self, rootWin, relief=appearance, bd=bd)
		self.canvas = tix.Canvas(
			self,
			height=height,
			width=width,
			bd=0,
			highlightthickness=0,
			background=background,
		)
		self.scale = self.canvas.create_rectangle(
			0,
			0,
			width,
			height,
			fill=fillColor,
		)
		self.label = self.canvas.create_text(
			width/2,
			height/2,
			text="",
			anchor="c",
			fill=labelColor,
			font=self.labelFont,
		)
		self.update()
		self.bind("<Configure>", self.update)
		self.canvas.pack(side="top", fill="x", expand="no")

	def updateProgress(self, newVal, newMax=None, text=""):
		if newMax:
			self.max = newMax
		self.value = newVal
		self.update(None, text)

	def update(self, event=None, labelText=""):
		# Trim the values to be between min and max
		value = self.value
		if value > self.max:
			value = self.max
		if value < self.min:
			value = self.min
		# Adjust the rectangle
		width = int(self.winfo_width())
		# width = self.width
		ratio = float(value)/self.max
		if self.orientation == "horizontal":
			self.canvas.coords(
				self.scale,
				0,
				0,
				width * ratio,
				self.height,
			)
		else:
			self.canvas.coords(
				self.scale,
				0,
				self.height * (1 - ratio),
				width,
				self.height,
			)
		# Now update the colors
		self.canvas.itemconfig(self.scale, fill=self.fillColor)
		self.canvas.itemconfig(self.label, fill=self.labelColor)
		# And update the label
		if not labelText:
			labelText = self.labelFormat % int(ratio * 100)
		self.canvas.itemconfig(self.label, text=labelText)
		# FIXME:
		# self.canvas.move(self.label, width/2, self.height/2)
		# self.canvas.scale(self.label, 0, 0, float(width)/self.width, 1)
		self.canvas.update_idletasks()



class FormatOptionsButton(ttk.Button):
	def __init__(
		self,
		kind: Union["Read", "Write"],
		values: Dict,
		formatVar: tk.StringVar,
		master=None,
	):
		ttk.Button.__init__(
			self,
			master=master,
			text="Options",
			command=self.buttonClicked,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		self.kind = kind
		self.kindFormatsOptions = {
			"Read": Glossary.formatsReadOptions,
			"Write": Glossary.formatsWriteOptions,
		}
		self.values = values
		self.formatVar = formatVar
		self.menu = None

	def valueMenuItemCustomSelected(self, treev, format, optName, menu=None):
		if menu:
			menu.destroy()
			self.menu = None

		value = treev.set(optName, self.valueCol)

		dialog = tix.Toplevel(master=treev) # bg="#0f0" does not work
		dialog.resizable(width=True, height=True)
		dialog.title(optName)
		set_window_icon(dialog)
		dialog.bind('<Escape>', lambda e: dialog.destroy())

		px, py, pw, ph = decodeGeometry(treev.winfo_toplevel().geometry())
		w = 300
		h = 100
		dialog.geometry(encodeGeometry(
			px + pw//2 - w//2,
			py + ph//2 - h//2,
			w,
			h,
		))
		
		frame = tix.Frame(master=dialog)

		label = ttk.Label(master=frame, text="Value for " + optName)
		label.pack()

		entry = ttk.Entry(master=frame)
		entry.insert(0, value)
		entry.pack(fill="x")

		prop = Glossary.formatsOptionsProp[format][optName]

		def okClicked(event=None):
			rawValue = entry.get()
			if not prop.validateRaw(rawValue):
				log.error(f"invalid {prop.typ} value: {optName} = {rawValue!r}")
				return
			treev.set(optName, self.valueCol, rawValue)
			treev.set(optName, "#1", "1") # enable it
			col_w = tkFont.Font().measure(rawValue)
			if treev.column("Value", width=None) < col_w:
				treev.column("Value", width=col_w)
			dialog.destroy()

		entry.bind("<Return>", okClicked)

		label = ttk.Label(master=frame)
		label.pack(fill="x")

		button = ttk.Button(
			frame,
			text="Ok",
			command=okClicked,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		button.pack(side="right")
		###
		frame.pack(fill="x")
		dialog.focus()

	def buttonClicked(self):
		formatD = self.formatVar.get()
		if not formatD:
			return
		format = Glossary.descFormat[formatD]
		options = self.kindFormatsOptions[self.kind][format]
		optionsProp = Glossary.formatsOptionsProp[format]

		dialog = tix.Toplevel()  # bg="#0f0" does not work
		dialog.resizable(width=True, height=True)
		dialog.title(self.kind + " Options")
		set_window_icon(dialog)
		dialog.bind('<Escape>', lambda e: dialog.destroy())
		###
		self.valueCol = "#3"
		cols = [
			"Enable", # bool
			"Name", # str
			"Value", # str
			"Comment", # str
		]
		treev = ttk.Treeview(
			master=dialog,
			columns=cols,
			show="headings",
		)
		for col in cols:
			treev.heading(
				col,
				text=col,
				#command=lambda c=col: sortby(treev, c, 0),
			)
			# adjust the column's width to the header string
			treev.column(
				col,
				width=tkFont.Font().measure(col.title()),
			)
		###
		def valueMenuItemSelected(optName, menu, value):
			treev.set(optName, self.valueCol, value)
			treev.set(optName, "#1", "1") # enable it
			col_w = tkFont.Font().measure(value)
			if treev.column("Value", width=None) < col_w:
				treev.column("Value", width=col_w)
			menu.destroy()
			self.menu = None
		def valueCellClicked(event, optName):
			if not optName:
				return
			prop = optionsProp[optName]
			propValues = prop.values
			if not propValues:
				if prop.customValue:
					self.valueMenuItemCustomSelected(treev, format, optName, None)
				else:
					log.error("invalid option %s, values=%s, customValue=%s" % (propValues, prop.customValue))
				return
			if prop.typ == "bool":
				rawValue = treev.set(optName, self.valueCol)
				if rawValue == "":
					value = False
				else:
					value, isValid = prop.evaluate(rawValue)
					if not isValid:
						log.error("invalid %s = %r" % (optName, rawValue))
						value = False
				treev.set(optName, self.valueCol, str(not value))
				treev.set(optName, "#1", "1") # enable it
				return
			menu = tk.Menu(
				master=treev,
				title=optName,
				tearoff=False,
			)
			self.menu = menu # to destroy it later
			if prop.customValue:
				menu.add_command(
					label="[Custom Value]",
					command=lambda: self.valueMenuItemCustomSelected(treev, format, optName, menu),
				)
			groupedValues = None
			if len(propValues) > 10:
				groupedValues = prop.groupValues()
			maxItemW = 0
			if groupedValues:
				for groupName, subValues in groupedValues.items():
					if subValues is None:
						menu.add_command(
							label=str(value),
							command=lambda value=value: valueMenuItemSelected(optName, menu, value),
						)
						maxItemW = max(maxItemW, tkFont.Font().measure(str(value)))
					else:
						subMenu = tk.Menu(tearoff=False)
						for subValue in subValues:
							subMenu.add_command(
								label=str(subValue),
								command=lambda value=subValue: valueMenuItemSelected(optName, menu, value),
							)
						menu.add_cascade(label=groupName, menu=subMenu)
						maxItemW = max(maxItemW, tkFont.Font().measure(groupName))
			else:
				for value in propValues:
					value = str(value)
					menu.add_command(
						label=value,
						command=lambda value=value: valueMenuItemSelected(optName, menu, value),
					)
			def close():
				menu.destroy()
				self.menu = None
			menu.add_command(
				label="[Close]",
				command=close,
			)
			try:
				menu.tk_popup(
					event.x_root,
					event.y_root,
				)
				# do not pass the third argument (entry), so that the menu
				# apears where the pointer is on its top-left corner
			finally:
				# make sure to release the grab (Tk 8.0a1 only)
				menu.grab_release()
		def treeClicked(event):
			if self.menu:
				self.menu.destroy()
				self.menu = None
				return
			optName = treev.identify_row(event.y) # optName is rowId
			if not optName:
				return
			col = treev.identify_column(event.x) # "#1" to self.valueCol
			if col == "#1":
				value = treev.set(optName, col)
				treev.set(optName, col, 1-int(value))
				return
			if col == self.valueCol:
				valueCellClicked(event, optName)
		treev.bind(
			"<Button-1>",
			# "<<TreeviewSelect>>", # event.x and event.y are zero
			treeClicked,
		)
		treev.pack(fill="x", expand=True)
		###
		for optName in options:
			prop = optionsProp[optName]
			comment = prop.typ
			if prop.comment:
				comment += ", " + prop.comment
			row = [
				int(optName in self.values),
				optName,
				str(self.values.get(optName, "")),
				comment,
			]
			treev.insert("", "end", values=row, iid=optName) # iid should be rowId
			# adjust column's width if necessary to fit each value
			for col_i, value in enumerate(row):
				value = str(value)
				if col_i == 3:
					value = value.zfill(20) # to reserve window width, because it's hard to resize it later
				col_w = tkFont.Font().measure(value)
				if treev.column(cols[col_i], width=None) < col_w:
					treev.column(cols[col_i], width=col_w)
		###########
		frame = tix.Frame(dialog)
		###
		def okClicked():
			for optName in options:
				enable = bool(int(treev.set(optName, "#1")))
				if not enable:
					if optName in self.values:
						del self.values[optName]
					continue
				rawValue = treev.set(optName, self.valueCol)
				prop = optionsProp[optName]
				value, isValid = prop.evaluate(rawValue)
				if not isValid:
					log.error("invalid option value %s = %s" % (optName, rawValue))
					continue
				self.values[optName] = value
			dialog.destroy()
		button = ttk.Button(
			frame,
			text="OK",
			command=okClicked,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		button.pack(side="right")
		###
		frame.pack(fill="x")
		###
		# x, y, w, h = decodeGeometry(dialog.geometry())
		w, h = 380, 250
		# w and h are rough estimated width and height of `dialog`
		px, py, pw, ph = decodeGeometry(self.winfo_toplevel().geometry())
		# move dialog without changing the size
		dialog.geometry(encodeLocation(
			px + pw//2 - w//2,
			py + ph//2 - h//2,
		))
		dialog.focus()


class UI(tix.Frame, UIBase):
	def __init__(self, path="", **options):
		self.glos = Glossary(ui=self)
		self.pref = {}
		self.pref_load(**options)
		#############################################
		rootWin = self.rootWin = tix.Tk()
		# a hack that hides the window until we move it to the center of screen
		if os.sep == "\\": # Windows
			rootWin.attributes('-alpha', 0.0)
		else: # Linux
			rootWin.withdraw()
		tix.Frame.__init__(self, rootWin)
		rootWin.title("PyGlossary (Tkinter)")
		rootWin.resizable(True, False)
		########
		set_window_icon(rootWin)
		rootWin.bind('<Escape>', lambda e: rootWin.quit())
		########
		self.pack(fill="x")
		# rootWin.bind("<Configure>", self.resized)
		######################
		self.glos = Glossary(ui=self)
		self.pref = {}
		self.pref_load()
		self.pathI = ""
		self.pathO = ""
		self.fcd_dir = join(homeDir, "Desktop")
		######################
		vpaned = ttk.PanedWindow(self, orient=tk.VERTICAL)
		notebook = tix.NoteBook(vpaned)
		notebook.add("tab1", label="Convert", underline=0)
		notebook.add("tab2", label="Reverse", underline=0)
		convertFrame = tix.Frame(notebook.tab1)
		######################
		frame = tix.Frame(convertFrame)
		##
		label = ttk.Label(frame, text="Read from format")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = ttk.OptionMenu(
			frame,
			comboVar,
			None, # default
			*Glossary.readDesc,
		)
		combo.pack(side="left")
		comboVar.trace("w", self.inputComboChanged)
		self.combobox_i = comboVar
		##
		self.readOptions = {} # type: Dict[str, Any]
		self.writeOptions = {} # type: Dict[str, Any]
		##
		self.readOptionsButton = FormatOptionsButton(
			"Read",
			self.readOptions,
			self.combobox_i,
			master=frame,
		)
		##
		frame.pack(fill="x")
		###################
		frame = tix.Frame(convertFrame)
		##
		label = ttk.Label(frame, text="  Path:")
		label.pack(side="left")
		##
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		entry.bind_all("<KeyPress>", self.entry_changed)
		self.entry_i = entry
		##
		button = ttk.Button(
			frame,
			text="Browse",
			command=self.browse_i,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		frame.pack(fill="x")
		######################
		frame = tix.Frame(convertFrame)
		##
		label = ttk.Label(frame, text="Write to format    ")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = ttk.OptionMenu(
			frame,
			comboVar,
			None, # default
			*Glossary.writeDesc,
		)
		combo.pack(side="left")
		comboVar.trace("w", self.outputComboChanged)
		self.combobox_o = comboVar
		##
		self.writeOptionsButton = FormatOptionsButton(
			"Write",
			self.writeOptions,
			self.combobox_o,
			master=frame,
		)
		##
		frame.pack(fill="x")
		###################
		frame = tix.Frame(convertFrame)
		##
		label = ttk.Label(frame, text="  Path:")
		label.pack(side="left")
		##
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		# entry.bind_all("<KeyPress>", self.entry_changed)
		self.entry_o = entry
		##
		button = ttk.Button(
			frame,
			text="Browse",
			command=self.browse_o,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		frame.pack(fill="x")
		#######
		frame = tix.Frame(convertFrame)
		label = ttk.Label(frame, text=" "*15)
		label.pack(
			side="left",
			fill="x",
			expand=True,
		)
		button = ttk.Button(
			frame,
			text="Convert",
			command=self.convert,
			# bg="#00e000",
			# activebackground="#22f022",
		)
		button.pack(
			side="left",
			fill="x",
			expand=True,
		)
		###
		frame.pack(fill="x")
		######
		convertFrame.pack(fill="x")
		vpaned.add(notebook)
		#################
		console = tix.Text(vpaned, height=15, background="#000000")
		# self.consoleH = 15
		# sbar = Tix.Scrollbar(
		#	vpaned,
		#	orien=Tix.VERTICAL,
		#	command=console.yview
		# )
		# sbar.grid ( row=0, column=1)
		# console["yscrollcommand"] = sbar.set
		# console.grid()
		console.pack(fill="both", expand=True)
		log.addHandler(
			TkTextLogHandler(console),
		)
		console.insert("end", "Console:\n")
		####
		vpaned.add(console)
		vpaned.pack(fill="both", expand=True)
		self.console = console
		##################
		frame2 = tix.Frame(self)
		clearB = ttk.Button(
			frame2,
			text="Clear",
			command=self.console_clear,
			# bg="black",
			# fg="#ffff00",
			# activebackground="#333333",
			# activeforeground="#ffff00",
		)
		clearB.pack(side="left")
		####
		label = ttk.Label(frame2, text="Verbosity")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = ttk.OptionMenu(
			frame2,
			comboVar,
			log.getVerbosity(), # default
			0, 1, 2, 3, 4,
		)
		comboVar.trace("w", self.verbosityChanged)
		combo.pack(side="left")
		self.verbosityCombo = comboVar
		#####
		pbar = ProgressBar(frame2, width=400)
		pbar.pack(side="left", fill="x", expand=True)
		self.pbar = pbar
		frame2.pack(fill="x")
		self.progressTitle = ""
		#############
		# vpaned.grid()
		# bottomFrame.grid()
		# self.grid()
		#####################
		# lbox = Tix.Listbox(convertFrame)
		# lbox.insert(0, "aaaaaaaa", "bbbbbbbbbbbbbbbbbbbb")
		# lbox.pack(fill="x")
		##############
		frame3 = tix.Frame(self)
		aboutB = ttk.Button(
			frame3,
			text="About",
			command=self.about_clicked,
			# bg="#e000e0",
			# activebackground="#f030f0",
		)
		aboutB.pack(side="right")
		closeB = ttk.Button(
			frame3,
			text="Close",
			command=rootWin.quit,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		closeB.pack(side="right")
		frame3.pack(fill="x")
		# __________________________ Reverse Tab __________________________ #
		revFrame = tix.Frame(notebook.tab2)
		revFrame.pack(fill="x")
		######################
		frame = tix.Frame(revFrame)
		##
		label = ttk.Label(frame, text="Read from format")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = ttk.OptionMenu(
			frame,
			comboVar,
			None, # default
			*Glossary.readDesc,
		)
		combo.pack(side="left")
		self.combobox_r_i = comboVar
		##
		frame.pack(fill="x")
		###################
		frame = tix.Frame(revFrame)
		##
		label = ttk.Label(frame, text="  Path:")
		label.pack(side="left")
		##
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		# entry.bind_all("<KeyPress>", self.entry_r_i_changed)
		self.entry_r_i = entry
		##
		button = ttk.Button(
			frame,
			text="Browse",
			command=self.r_browse_i,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		button = ttk.Button(
			frame,
			text="Load",
			command=self.r_load,
			# bg="#7777ff",
		)
		button.pack(side="left")
		###
		frame.pack(fill="x")
		###################
		frame = tix.Frame(revFrame)
		##
		label = ttk.Label(frame, text="Output Tabfile")
		label.pack(side="left")
		###
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		# entry.bind_all("<KeyPress>", self.entry_r_i_changed)
		self.entry_r_o = entry
		##
		button = ttk.Button(
			frame,
			text="Browse",
			command=self.r_browse_o,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		frame.pack(fill="x")
		#############
		centerWindow(rootWin)
		# show the window again
		if os.sep == "\\": # Windows
			rootWin.attributes('-alpha', 1.0)
		else: # Linux
			rootWin.deiconify()
		##############################
		if path:
			self.entry_i.insert(0, path)
			self.entry_changed()
			self.load()

	def verbosityChanged(self, index, value, op):
		log.setVerbosity(
			int(self.verbosityCombo.get())
		)

	def about_clicked(self):
		about = tix.Toplevel(width=600)  # bg="#0f0" does not work
		about.title("About PyGlossary")
		about.resizable(width=False, height=False)
		set_window_icon(about)
		about.bind('<Escape>', lambda e: about.destroy())
		###
		msg1 = tix.Message(
			about,
			width=350,
			text="PyGlossary %s (Tkinter)" % VERSION,
			font=("DejaVu Sans", 13, "bold"),
		)
		msg1.pack(fill="x", expand=True)
		###
		msg2 = tix.Message(
			about,
			width=350,
			text=aboutText,
			font=("DejaVu Sans", 9, "bold"),
			justify=tix.CENTER,
		)
		msg2.pack(fill="x", expand=True)
		###
		msg3 = tix.Message(
			about,
			width=350,
			text=homePage,
			font=("DejaVu Sans", 8, "bold"),
			fg="#3333ff",
		)
		msg3.pack(fill="x", expand=True)
		###
		msg4 = tix.Message(
			about,
			width=350,
			text="Install Gtk3+PyGI to have a better interface!",
			font=("DejaVu Sans", 8, "bold"),
			fg="#00aa00",
		)
		msg4.pack(fill="x", expand=True)
		###########
		frame = tix.Frame(about)
		###
		button = ttk.Button(
			frame,
			text="Close",
			command=about.destroy,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		button.pack(side="right")
		###
		button = ttk.Button(
			frame,
			text="License",
			command=self.about_license_clicked,
			# bg="#00e000",
			# activebackground="#22f022",
		)
		button.pack(side="right")
		###
		button = ttk.Button(
			frame,
			text="Credits",
			command=self.about_credits_clicked,
			# bg="#0000ff",
			# activebackground="#5050ff",
		)
		button.pack(side="right")
		###
		frame.pack(fill="x")

	def about_credits_clicked(self):
		about = tix.Toplevel()  # bg="#0f0" does not work
		about.title("Credits")
		about.resizable(width=False, height=False)
		set_window_icon(about)
		about.bind('<Escape>', lambda e: about.destroy())
		###
		msg1 = tix.Message(
			about,
			width=500,
			text="\n".join(authors),
			font=("DejaVu Sans", 9, "bold"),
		)
		msg1.pack(fill="x", expand=True)
		###########
		frame = tix.Frame(about)
		closeB = ttk.Button(
			frame,
			text="Close",
			command=about.destroy,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		closeB.pack(side="right")
		frame.pack(fill="x")

	def about_license_clicked(self):
		about = tix.Toplevel()  # bg="#0f0" does not work
		about.title("License")
		about.resizable(width=False, height=False)
		set_window_icon(about)
		about.bind('<Escape>', lambda e: about.destroy())
		###
		msg1 = tix.Message(
			about,
			width=420,
			text=licenseText,
			font=("DejaVu Sans", 9, "bold"),
		)
		msg1.pack(fill="x", expand=True)
		###########
		frame = tix.Frame(about)
		closeB = ttk.Button(
			frame,
			text="Close",
			command=about.destroy,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		closeB.pack(side="right")
		frame.pack(fill="x")

	def resized(self, event):
		dh = self.rootWin.winfo_height() - self.winfo_height()
		# log.debug(dh, self.consoleH)
		# if dh > 20:
		#	self.consoleH += 1
		#	self.console["height"] = self.consoleH
		#	self.console["width"] = int(self.console["width"]) + 1
		#	self.console.grid()
		# for x in dir(self):
		#	if "info" in x:
		#		log.debug(x)

	def inputComboChanged(self, *args):
		formatD = self.combobox_i.get()
		if not formatD:
			return
		self.readOptions.clear() # reset the options, DO NOT re-assign
		format = Glossary.descFormat[formatD]
		options = Glossary.formatsReadOptions[format]
		if options:
			self.readOptionsButton.pack(side="right")
		else:
			self.readOptionsButton.pack_forget()

	def outputComboChanged(self, *args):
		# log.debug(self.combobox_o.get())
		formatD = self.combobox_o.get()
		if not formatD:
			return
		self.writeOptions.clear() # reset the options, DO NOT re-assign
		format = Glossary.descFormat[formatD]
		options = Glossary.formatsWriteOptions[format]
		if options:
			self.writeOptionsButton.pack(side="right")
		else:
			self.writeOptionsButton.pack_forget()

		if not self.pref["ui_autoSetOutputFileName"]:  # and format is None:
			return

		pathI = self.entry_i.get()
		pathO = self.entry_o.get()
		formatOD = self.combobox_o.get()

		if formatOD is None:
			return
		if pathO:
			return
		if "." not in pathI:
			return

		extO = Glossary.descExt[formatOD]
		pathO = "".join(os.path.splitext(pathI)[:-1])+extO
		# self.entry_o.delete(0, "end")
		self.entry_o.insert(0, pathO)

	def entry_changed(self, event=None):
		# log.debug("entry_changed")
		# char = event.keysym
		pathI = self.entry_i.get()
		if self.pathI != pathI:
			formatD = self.combobox_i.get()
			if pathI.startswith("file://"):
				pathI = urlToPath(pathI)
				self.entry_i.delete(0, "end")
				self.entry_i.insert(0, pathI)
			if self.pref["ui_autoSetFormat"]:
				ext = os.path.splitext(pathI)[-1].lower()
				if ext in (".gz", ".bz2", ".zip"):
					ext = os.path.splitext(pathI[:-len(ext)])[-1].lower()
				for i in range(len(Glossary.readExt)):
					if ext in Glossary.readExt[i]:
						self.combobox_i.set(Glossary.readDesc[i])
						break
			if self.pref["ui_autoSetOutputFileName"]:
				# pathI = self.entry_i.get()
				formatOD = self.combobox_o.get()
				pathO = self.entry_o.get()
				if formatOD and not pathO and "." in pathI:
					extO = Glossary.descExt[formatOD]
					pathO = "".join(os.path.splitext(pathI)[:-1]) + extO
					self.entry_o.delete(0, "end")
					self.entry_o.insert(0, pathO)
			self.pathI = pathI
		##############################################
		pathO = self.entry_o.get()
		if self.pathO != pathO:
			formatD = self.combobox_o.get()
			if pathO.startswith("file://"):
				pathO = urlToPath(pathO)
				self.entry_o.delete(0, "end")
				self.entry_o.insert(0, pathO)
			if self.pref["ui_autoSetFormat"]:
				ext = os.path.splitext(pathO)[-1].lower()
				if ext in (".gz", ".bz2", ".zip"):
					ext = os.path.splitext(pathO[:-len(ext)])[-1].lower()
				for i in range(len(Glossary.writeExt)):
					if ext in Glossary.writeExt[i]:
						self.combobox_o.set(Glossary.writeDesc[i])
						break
			self.pathO = pathO

	def browse_i(self):
		path = filedialog.askopenfilename(initialdir=self.fcd_dir)
		if path:
			self.entry_i.delete(0, "end")
			self.entry_i.insert(0, path)
			self.entry_changed()
			self.fcd_dir = os.path.dirname(path)  # FIXME

	def browse_o(self):
		path = filedialog.asksaveasfilename()
		if path:
			self.entry_o.delete(0, "end")
			self.entry_o.insert(0, path)
			self.entry_changed()
			self.fcd_dir = os.path.dirname(path)  # FIXME

	def convert(self):
		inPath = self.entry_i.get()
		if not inPath:
			log.critical("Input file path is empty!")
			return
		inFormatDesc = self.combobox_i.get()
		if not inFormatDesc:
			# log.critical("Input format is empty!");return
			inFormat = ""
		else:
			inFormat = Glossary.descFormat[inFormatDesc]

		outPath = self.entry_o.get()
		if not outPath:
			log.critical("Output file path is empty!")
			return
		outFormatDesc = self.combobox_o.get()
		if not outFormatDesc:
			log.critical("Output format is empty!")
			return
		outFormat = Glossary.descFormat[outFormatDesc]

		finalOutputFile = self.glos.convert(
			inPath,
			inputFormat=inFormat,
			outputFilename=outPath,
			outputFormat=outFormat,
			readOptions=self.readOptions,
			writeOptions=self.writeOptions,
		)
		# if finalOutputFile:
			# self.status("Convert finished")
		# else:
			# self.status("Convert failed")
		return bool(finalOutputFile)

	def run(self, editPath=None, readOptions=None):
		if readOptions is None:
			readOptions = {}
		# editPath and readOptions are for DB Editor
		# which is not implemented
		self.mainloop()

	def progressInit(self, title):
		self.progressTitle = title

	def progress(self, rat, text=""):
		if not text:
			text = "%%%d" % (rat*100)
		text += " - %s" % self.progressTitle
		self.pbar.updateProgress(rat*100, None, text)
		# self.pbar.value = rat*100
		# self.pbar.update()
		self.rootWin.update()

	def console_clear(self, event=None):
		self.console.delete("1.0", "end")
		self.console.insert("end", "Console:\n")

	def r_browse_i(self):
		pass

	def r_browse_o(self):
		pass

	def r_load(self):
		pass


if __name__ == "__main__":
	import sys
	if len(sys.argv) > 1:
		path = sys.argv[1]
	else:
		path = ""
	ui = UI(path)
	ui.run()



# -*- coding: utf-8 -*-

from time import time as now
import re

from formats_common import *

enable = True
format = "WikipediaDump"
description = "Wikipedia Dump(Static HTML)"
extensions = [".wiki"]
optionsProp = {
	"encoding": EncodingOption(),
}
depends = {}

BeautifulSoup = None

def loadBeautifulSoup():
	global BeautifulSoup
	try:
		import bs4 as BeautifulSoup
	except:
		import BeautifulSoup
	if int(BeautifulSoup.__version__.split(".")[0]) < 4:
		raise ImportError(
			"BeautifulSoup is too old, required at least version 4, " +
			"%r found.\n" % BeautifulSoup.__version__ +
			"Please run `sudo pip3 install lxml beautifulsoup4 html5lib`"
		)

class Reader(object):
	specialPattern = re.compile(r".*~[^_]")
	def __init__(self, glos):
		self._glos = glos
		self._rootDir = ""
		self._articlesDir = ""
		self._len = None
		self._specialCount = 0
		# self._alts = {}
		# { word => alts }
		# where alts is str (one word), or list of strs
		# we can't recognize alternates unless we keep all data in memory
		# or scan the whole directiry and read all files twice

	def open(self, dirname):
		if not isdir(dirname):
			raise IOError("%s is not a directory" % dirname)
		self._rootDir = dirname
		self._articlesDir = join(self._rootDir, "articles")

	def close(self):
		self._rootDir = ""
		self._articlesDir = ""
		self._len = None
		# self._alts = {}

	def __len__(self):
		if not self._articlesDir:
			log.error(
				"WikipediaDump: called len(reader) while it's not open"
			)
			return 0
		if self._len is None:
			t0 = now()
			log.info("Counting articles...")
			self._len = sum(
				len(files)
				for _, _, files in os.walk(self._articlesDir)
			)
			log.debug("Counting articles took %.2f seconds", now() - t0)
			log.info("Found %s articles", self._len)
		return self._len

	def __iter__(self):
		if not self._articlesDir:
			log.error(
				"WikipediaDump: trying to iterate over reader"
				" while it's not open"
			)
			raise StopIteration
		for dirpath, dirs, files in os.walk(self._articlesDir):
			# dirpathRel = dirpath[len(self._articlesDir):].lstrip("/")
			# dirParts = dirpathRel.split(os.sep)  # test on windows FIXME
			# prefix = "".join([
			#	chr(int(x, 16)) if len(x)==2 else x
			#	for x in dirParts
			# ])
			dirs.sort()
			files.sort()
			for fname_html in files:
				fpath = join(dirpath, fname_html)
				fname, ext = splitext(fname_html)
				# fpathRel = join(dirpathRel, fname_html)
				if ext != ".html":
					log.warning("unkown article extension: %s", fname_ext)
					continue
				if self.isSpecialByPath(fname):  # , prefix
					# log.debug("Skipping special page file: %s", fpathRel)
					self._specialCount += 1
					yield None  # updates progressbar
					continue
				word = fname.replace("_", " ").replace("~", ":")
				defi = self.parseArticle(word, fpath)
				if not defi:
					yield None  # updates progressbar
					continue

				yield self._glos.newEntry(word, defi)
		log.info("Skipped %s special page files", self._specialCount)

	def isSpecialByPath(self, fname):  # , d_prefix
		"""
			fname: str
			d_prefix: str, with length of 3
				this is the joint string version of directory relative path
		"""
		return re.match(self.specialPattern, fname)
		# assert len(d_prefix) == 3
		# f_prefix = fname[:3].lower()
		# if f_prefix == d_prefix:
		#	return False
		# if len(f_prefix) < 3:
		#	if f_prefix == d_prefix.rstrip("_"):
		#		return False
		# return True
		# if "~" not in fname:
		#	return False
		# if fname[0] == dirParts[0]:
		#	return False
		# if list(fname[:3]) != dirParts:
		#	log.debug("dirParts=%r, fname=%r", dirParts, fname)
		# l_fname = fname.lower()
		# for ext in ("png", "jpg", "jpeg", "gif", "svg", "pdf", "js"):
		#	if "." + ext in l_fname:
		#		return True
		# return True

	def parseArticle(self, word, fpath):
		if BeautifulSoup is None:
			loadBeautifulSoup()
		try:
			with open(fpath) as fileObj:
				text = fileObj.read()
		except UnicodeDecodeError:
			log.error("error decoding file %r, not UTF-8", fpath)
			return
		except:
			log.exception("error reading file %r", fpath)
			return

		root = BeautifulSoup(text, "lxml")
		body = root.body
		if not body:
			return

#		if body.p and body.p.text.startswith("Redirecting to "):
#			toWord = body.p.text[len("Redirecting to "):]
#			try:
#				fromWords = self._alts[toWord]
#			except KeyError:
#				self._alts[toWord] = word
#			else:
#				if isinstance(fromWords, str):
#					self._alts[toWord] = [fromWords, word]
#				else:
#					fromWords.append(word)
#			return
#		try:
#			alts = self._alts[word]

		if body.p and body.p.text.startswith("Redirecting to "):
			toWord = body.p.text[len("Redirecting to "):]
			return "↳ <a href=\"bword://%s\">%s</a>" % (toWord, toWord)
			# "↳" does not look good for RTL languages FIXME

		try:
			content = body.find(id="column-content").find(id="content")
		except AttributeError:
			content = None
		if not content:
			log.warning("could not find \"content\" element: %s", fpath)
			return

		try:
			firstHeading = content.find("h1", class_="firstHeading").text
		except AttributeError:
			log.warning("could not find \"firstHeading\" element: %s", fpath)
			return

		if firstHeading != word:
			log.debug("word=%r, firstHeading=%r", firstHeading, word)

		bodyContent = content.find(id="bodyContent")
		if not bodyContent:
			log.warning("could not find \"bodyContent\" element: %s", fpath)
			return

		# FIXME
		return "".join([str(tag) for tag in bodyContent.contents])



# -*- coding: utf-8 -*-
import re
from string import Template

from tabfile import Reader as TabfileReader
from formats_common import *

enable = True
format = "Dicformids"
description = "DictionaryForMIDs"
extensions = [".mids"]
optionsProp = {}
depends = {}


PROP_TEMPLATE = """#DictionaryForMIDs property file
infoText=$name, author: $author
indexFileMaxSize=$indexFileMaxSize\n
language1IndexNumberOfSourceEntries=%wordCount
language1DictionaryUpdateClassName=de.kugihan.dictionaryformids.dictgen.DictionaryUpdate
indexCharEncoding=ISO-8859-1
dictionaryFileSeparationCharacter='\\t'
language2NormationClassName=de.kugihan.dictionaryformids.translation.Normation
language2DictionaryUpdateClassName=de.kugihan.dictionaryformids.dictgen.DictionaryUpdate
logLevel=0
language1FilePostfix=$directoryPostfix
dictionaryCharEncoding=UTF-8
numberOfAvailableLanguages=2
language1IsSearchable=true
language2GenerateIndex=false
dictionaryFileMaxSize=$dicMaxSize
language2FilePostfix=$language2FilePostfix
searchListFileMaxSize=20000
language2IsSearchable=false
fileEncodingFormat=plain_format1
language1HasSeparateDictionaryFile=true
searchListCharEncoding=ISO-8859-1
searchListFileSeparationCharacter='\t'
indexFileSeparationCharacter='\t'
language1DisplayText=$inputlang
language2HasSeparateDictionaryFile=false
dictionaryGenerationInputCharEncoding=UTF-8
language1GenerateIndex=true
language2DisplayText=$outputlang
language1NormationClassName=de.kugihan.dictionaryformids.translation.NormationEng
"""


class Reader(object):
	def __init__(self, glos):
		self._glos = glos
		self._tabFileNames = []
		self._tabFileReader = None

	def open(self, dirname):
		self._dirname = dirname
		dicFiles = []
		orderFileNames = []
		for fname in os.listdir(dirname):
			if not fname.startswith("directory"):
				continue
			try:
				num = re.findall("\d+", fname)[-1]
			except IndexError:
				pass
			else:
				orderFileNames.append((num, fname))
		orderFileNames.sort(
			key=lambda x: x[0],
			reverse=True,
		)
		self._tabFileNames = [x[1] for x in orderFileNames]
		self.nextTabFile()

	def __len__(self):  # FIXME
		raise NotImplementedError

	def __iter__(self):
		return self

	def __next__(self):
		for _ in range(10):
			try:
				return next(self._tabFileReader)
			except StopIteration:
				self._tabFileReader.close()
				self.nextTabFile()

	def nextTabFile(self):
		try:
			tabFileName = self._tabFileNames.pop()
		except IndexError:
			raise StopIteration
		self._tabFileReader = TabfileReader(self._glos, hasInfo=False)
		self._tabFileReader.open(join(self._dirname, tabFileName))

	def close(self):
		if self._tabFileReader:
			try:
				self._tabFileReader.close()
			except:
				pass
		self._tabFileReader = None
		self._tabFileNames = []


class Writer(object):
	def __init__(self, glos):
		self._glos = glos
		self.linesPerDirectoryFile = 500  # 200
		self.indexFileMaxSize = 32722  # 30000
		self.directoryPostfix = ""
		self.indexPostfix = "Eng"
		self.dirname = ""

	def open(self, dirname):
		self.dirname = dirname
		if not os.path.isdir(dirname):
			os.mkdir(dirname)

	def writeGetIndexGen(self):
		dicMaxSize = 0
		wordCount = 0
		for dicIndex, entryList in enumerate(
			self._glos.iterEntryBuckets(
				self.linesPerDirectoryFile
			)
		):
			# assert len(entryList) == 200
			dicFp = open(join(
				self.dirname,
				"directory%s%d.csv" % (
					self.directoryPostfix,
					dicIndex+1,
				),
			), "w")
			for entry in entryList:
				if entry.isData():
					# FIXME
					continue

				wordCount += 1
				word = entry.getWord()
				defi = entry.getDefi()
				dicLine = "%s\t%s\n" % (word, defi)
				dicPos = dicFp.tell()
				dicFp.write(dicLine)
				yield word, dicIndex+1, dicPos

			dicMaxSize = max(dicMaxSize, dicFp.tell())
			dicFp.close()
		self.dicMaxSize = dicMaxSize
		self.wordCount = wordCount

	def writeProbs(self):
		glos = self._glos
		with open(join(
			self.dirname,
			"DictionaryForMIDs.properties",
		), "w") as fp:
			fp.write(Template(PROP_TEMPLATE).substitute(
				name=glos.getInfo("name"),
				author=glos.getInfo("author"),
				indexFileMaxSize=self.indexFileMaxSize,
				wordCount=self.wordCount,
				directoryPostfix=self.directoryPostfix,
				dicMaxSize=self.dicMaxSize+1,
				language2FilePostfix="fa",  # FIXME
				inputlang=glos.getInfo("inputlang"),
				outputlang=glos.getInfo("outputlang"),
			))
#			open(join(
#				self.dirname,
#				"searchlist%s.csv"%self.directoryPostfix
#			), "w")  # FIXME

	def nextIndex(self):
		try:
			self.indexFp.close()
		except AttributeError:
			self.indexIndex = 0

		self.indexIndex += 1
		fname = "index%s%d.csv" % (self.indexPostfix, self.indexIndex)
		fpath = join(self.dirname, fname)
		self.indexFp = open(fpath, "w")

	def write(self):
		self.nextIndex()
		for word, dicIndex, dicPos in self.writeGetIndexGen():
			indexLine = "%s\t%d-%d-B\n" % (
				word,
				dicIndex + 1,
				dicPos,
			)
			if (
				self.indexFp.tell() + len(indexLine)
			) > self.indexFileMaxSize - 10:
				self.nextIndex()
			self.indexFp.write(indexLine)
		self.indexFp.close()
		self.writeProbs()

	# def close(self):
	#	pass




# -*- coding: utf-8 -*-
# appledict/indexes/zh.py
#
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
"""
Chinese wildcard and pinyin indexes.
"""

import re

import bs4
import colorize_pinyin as color

from . import languages, log


def zh(titles, content):
	"""
	chinese indexes.

	assuming that content is HTML and pinyin is inside second tag
	(first is <h1>), we can try to parse pinyin and generate indexes
	with pinyin subwords separated by whitespaces
	- pinyin itself
	- pinyin with diacritics replaced by tone numbers

	multiple pronunciations separated by comma or semicolon are supported.
	"""

	indexes = set()

	for title in titles:
		# feature: put dot at the end to match only this word
		indexes.update({title, title + "。"})

		# remove all non hieroglyph
		title = re.sub(r"[^\u4e00-\u9fff]", "", title)
		indexes.add(title)

	indexes.update(pinyin_indexes(content))

	return indexes


def pinyin_indexes(content):
	pinyin = find_pinyin(content)
	# assert type(pinyin) == unicode

	if not pinyin or pinyin == "_":
		return ()

	indexes = set()

	# multiple pronunciations
	for pinyin in re.split(r",|;", pinyin):

		# find all pinyin ranges, use them to rip pinyin out
		py = [
			r._slice(pinyin)
			for r in color.ranges_of_pinyin_in_string(pinyin)
		]

		# maybe no pinyin here
		if not py:
			return ()

		# just pinyin, with diacritics, separated by whitespace
		indexes.add("%s." % color.utf(" ".join(py)))

		# pinyin with diacritics replaced by tone numbers
		indexes.add("%s." % color.utf(" ".join(
			["%s%d" % (
				color.lowercase_string_by_removing_pinyin_tones(p),
				color.determine_tone(p)) for p in py])))
	return indexes


def find_pinyin(content):
	# assume that content is HTML and pinyin is inside second tag
	# (first is <h1>)
	soup = bs4.BeautifulSoup(content.splitlines()[0], features="lxml")
	if soup.body:
		soup = soup.body
	children = soup.children
	try:
		next(children)
		pinyin = next(children)
	except StopIteration:
		return None
	return pinyin.text


languages["zh"] = zh



# -*- coding: utf-8 -*-
# The MIT License (MIT)

# Copyright (C) 2012-2016 Alberto Pettarin (alberto@albertopettarin.it)
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from formats_common import *
from pyglossary.ebook_base import *

enable = True
format = "Mobi"
description = "MOBI E-Book"
extensions = [".mobi",]
sortOnWrite = ALWAYS

optionsProp = {
	"group_by_prefix_length": IntOption(),
	# "group_by_prefix_merge_min_size": IntOption(),
	# "group_by_prefix_merge_across_first": BoolOption(),
	"no_kindlegen": BoolOption(), # specific to mobi
}



class Writer(EbookWriter):
	ebook_format = format

	CSS_CONTENTS = """"@charset "UTF-8";"""
	GROUP_XHTML_TEMPLATE = """<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>{title}</title>
  <link rel="stylesheet" type="text/css" href="style.css" />
 </head>
 <body id="groupPage" class="groupPage">
  <h1 class="groupTitle">{group_title}</h1>
  <div class="groupNavigation">
   <a href="{previous_link}">[ Previous ]</a>
{index_link}
   <a href="{next_link}">[ Next ]</a>
  </div>
{group_contents}
 </body>
</html>"""

	GROUP_XHTML_INDEX_LINK = """   <a href="index.xhtml">[ Index ]</a>"""

	GROUP_XHTML_WORD_TEMPLATE = """   <span class="groupHeadword"><idx:entry><idx:orth>{headword}</idx:orth></idx:entry></span>"""

	GROUP_XHTML_WORD_JOINER = " &#8226;\n"

	GROUP_XHTML_WORD_DEFINITION_TEMPLATE = """  <div class="groupEntry">
   <idx:entry>
	<h2 class="groupHeadword"><idx:orth>{headword}</idx:orth></h2>
	<p class="groupDefinition">{definition}</p>
   </idx:entry>
  </div>"""


	OPF_TEMPLATE = """<?xml version="1.0" encoding="utf-8"?>
<package unique-identifier="uid">
 <metadata>
  <dc-metadata xmlns:dc="http://purl.org/metadata/dublin_core" xmlns:oebpackage="http://openebook.org/namespaces/oeb-package/1.0/">
   <dc:Title>{title}</dc:Title>
   <dc:Language>{sourceLang}</dc:Language>
   <dc:Identifier id="uid">{identifier}</dc:Identifier>
   <dc:Creator>{creator}</dc:Creator>
   <dc:Rights>{copyright}</dc:Rights>
   <dc:Subject BASICCode="REF008000">Dictionaries</dc:Subject>
  </dc-metadata>
  <x-metadata>
   <output encoding="utf-8"></output>
   <DictionaryInLanguage>{sourceLang}</DictionaryInLanguage>
   <DictionaryOutLanguage>{targetLang}</DictionaryOutLanguage>
   <EmbeddedCover>{cover}</EmbeddedCover>
  </x-metadata>
 </metadata>
 <manifest>
{manifest}
 </manifest>
 <spine>
{spine}
 </spine>
 <tours></tours>
 <guide></guide>
</package>"""


	def __init__(self, glos, **kwargs):
		import uuid
		EbookWriter.__init__(
			self,
			glos,
		)
		glos.setInfo(
			"uuid",
			str(uuid.uuid4()).replace("-", ""),
		)

	def write(
		self,
		filename,
		no_kindlegen=False,
		group_by_prefix_length=2,
	):
		EbookWriter.write(
			self,
			filename,
			compress=False,
			keep=False,
			group_by_prefix_length=group_by_prefix_length,
			# mobi_no_kindlegen=no_kindlegen,
			# kindlegen_path=None,
		)
		self.close()

		"""
		# run kindlegen
		tmp_path = mobi.get_tmp_path()
		if no_kindlegen:
			print_info("Not running kindlegen, the raw files are located in "%s"" % tmp_path)
			result = [tmp_path]
		else:
			try:
				print_debug("Creating .mobi file with kindlegen...", args.debug)
				kindlegen_path = KINDLEGEN
				opf_file_path_absolute = os.path.join(tmp_path, "OEBPS", "content.opf")
				mobi_file_path_relative = "content.mobi"
				mobi_file_path_absolute = os.path.join(tmp_path, "OEBPS", mobi_file_path_relative)
				if args.kindlegen_path is None:
					print_info("  Running '%s' from $PATH" % KINDLEGEN)
				else:
					kindlegen_path = args.kindlegen_path
					print_info("  Running '%s' from '%s'" % (KINDLEGEN, kindlegen_path))
				proc = subprocess.Popen(
					[kindlegen_path, opf_file_path_absolute, "-o", mobi_file_path_relative],
					stdout=subprocess.PIPE,
					stdin=subprocess.PIPE,
					stderr=subprocess.PIPE
				)
				output = proc.communicate()
				if args.debug:
					output_unicode = (output[0]).decode("utf-8")
					print_debug(output_unicode, args.debug)
				copy_file(mobi_file_path_absolute, output_file_path_absolute)
				result = [output_file_path]
				print_debug("Creating .mobi file with kindlegen... done", args.debug)
			except OSError as exc:
				print_error("  Unable to run '%s' as '%s'" % (KINDLEGEN, kindlegen_path))
				print_error("  Please make sure '%s':" % KINDLEGEN)
				print_error("    1. is available on your $PATH or")
				print_error("    2. specify its path with --kindlegen-path")
		"""
		return True




# -*- coding: utf-8 -*-

from formats_common import *
from pyglossary.text_reader import TextGlossaryReader
from pyglossary.file_utils import fileCountLines

enable = True
format = "LingoesLDF"
description = "Lingoes Source (LDF)"
extensions = [".ldf"]
optionsProp = {
	"newline": NewlineOption(),
	"resources": BoolOption(),
	"encoding": EncodingOption(),
}
depends = {}

infoKeys = [
	"title",
	"description",
	"author",
	"email",
	"website",
	"copyright",
]


class Reader(TextGlossaryReader):
	def __len__(self):
		if self._wordCount is None:
			log.debug("Try not to use len(reader) as it takes extra time")
			self._wordCount = fileCountLines(
				self._filename,
				newline="\n\n",
			) - self._leadingLinesCount
		return self._wordCount

	def isInfoWord(self, word):
		if isinstance(word, str):
			return word.startswith("#")
		else:
			return False

	def fixInfoWord(self, word):
		if isinstance(word, str):
			return word.lstrip("#")
		else:
			return word

	def loadInfo(self): # FIXME
		pass

	def nextPair(self):
		if not self._file:
			raise StopIteration
		entryLines = []
		while True:
			line = self._file.readline()
			if not line:
				raise StopIteration
			line = line.rstrip("\n\r")  # FIXME
			if line:
				entryLines.append(line)
				continue

			# now `line` is empty, process `entryLines`
			if not entryLines:
				return
			if len(entryLines) < 2:
				log.error(
					"invalid block near line %s" % fileObj.line +
					" in file %s" % filename
				)
				return
			word = entryLines[0]
			defi = "\n".join(entryLines[1:])
			defi = defi.replace("<br/>", "\n")  # FIXME

			word = [p.strip() for p in word.split("|")]

			return word, defi


def read(glos: GlossaryType, filename: str):
	glos.setDefaultDefiFormat("h")
	fileObj = FileLineWrapper(open(filename))
	entryLines = []

	def addDataEntry(entryLines):
		if not entryLines:
			return
		if len(entryLines) < 2:
			log.error(
				"invalid block near line %s" % fileObj.line +
				" in file %s" % filename
			)
			return
		word = entryLines[0]
		defi = "\n".join(entryLines[1:])
		defi = defi.replace("<br/>", "\n")  # FIXME

		word = [p.strip() for p in word.split("|")]

		glos.addEntry(
			word,
			defi,
		)

	for line in fileObj:
		line = line.strip()
		if not line.startswith("###"):
			if line:
				entryLines.append(line)
			break
		parts = line[3:].split(":")
		if not parts:
			continue
		key = parts[0].lower()
		value = " ".join(parts[1:]).strip()
		glos.setInfo(key, value)
	# info lines finished

	for line in fileObj:
		line = line.strip()
		if line:
			entryLines.append(line)
		else:
			addDataEntry(entryLines)
			entryLines = []

	addDataEntry(entryLines)


def write(
	glos: GlossaryType,
	filename: str,
	newline: str = "\n",
	resources: str = True,
):
	g = glos
	head = "\n".join([
		"###%s: %s" % (
			key.capitalize(),
			g.getInfo(key),
		)
		for key in infoKeys
	])
	head += "\n"
	g.writeTxt(
		"\n",
		"\n\n",
		filename=filename,
		writeInfo=False,
		rplList=(
			("\n", "<br/>"),
		),
		ext=".ldf",
		head=head,
		newline=newline,
		resources=resources,
	)



#!/usr/bin/env python3

try:
	import py2exe
except ImportError:
	py2exe = None

import glob
import sys
import os
from os.path import join, dirname, exists, isdir
import re
import logging

import setuptools
from distutils.core import setup
from distutils.command.install import install

from pyglossary.glossary import VERSION

log = logging.getLogger("root")
relRootDir = "share/pyglossary"


class my_install(install):
	def run(self):
		install.run(self)
		if os.sep == "/":
			binPath = join(self.install_scripts, "pyglossary")
			log.info("creating script file \"%s\"", binPath)
			if not exists(self.install_scripts):
				os.makedirs(self.install_scripts)
				# let it fail on wrong permissions.
			else:
				if not isdir(self.install_scripts):
					raise OSError(
						"installation path already exists " +
						"but is not a directory: %s" % self.install_scripts
					)
			open(binPath, "w").write(
				join(self.install_data, relRootDir, "main.py") +
				" \"$@\""  # pass options from command line
			)
			os.chmod(binPath, 0o755)


data_files = [
	(relRootDir, [
		"about",
		"license.txt",
		"license-dialog",
		"help",
		"main.py",
		"pyglossary.pyw",
		"AUTHORS",
		"config.json",
	]),
	(relRootDir+"/ui", glob.glob("ui/*.py")),
	(relRootDir+"/ui/glade", glob.glob("ui/glade/*")),
	(relRootDir+"/ui/gtk3_utils", glob.glob("ui/gtk3_utils/*.py")),
	(relRootDir+"/res", glob.glob("res/*")),
	("share/doc/pyglossary", []),
	("share/doc/pyglossary/non-gui_examples",
		glob.glob("doc/non-gui_examples/*")),
	("share/doc/pyglossary/Babylon", glob.glob("doc/Babylon/*")),
	("share/doc/pyglossary/DSL", glob.glob("doc/DSL/*")),
	("share/doc/pyglossary/Octopus MDict", glob.glob("doc/Octopus MDict/*")),
	("share/applications", ["pyglossary.desktop"]),
	("share/pixmaps", ["res/pyglossary.png"]),
]


def files(folder):
	for path in glob.glob(folder+"/*"):
		if os.path.isfile(path):
			yield path

if py2exe:
	py2exeoptions = {
		"script_args": ["py2exe"],
		"bundle_files": 1,
		"windows": [
			{
				"script": "main.py", # or pyglossary.pyw ?
				"icon_resources": [
					(1, "res/pyglossary.ico"),
				],
			},
		],
		"zipfile": None,
		"options": {
			"py2exe": {
				"packages": [
					"pyglossary",
					"ui.ui_tk",
					# "ui.ui_gtk",
					# "BeautifulSoup",
					# "xml",
					"Tkinter", "tkFileDialog", "Tix",
					# "gtk", "glib", "gobject",
				],
			},
		},
	}
	data_files = [
		("tcl/tix8.1", files(join(sys.prefix, "tcl", "tix8.4.3"))),
		("tcl/tix8.1/bitmaps",
			files(join(sys.prefix, "tcl", "tix8.4.3", "bitmaps"))),
		("tcl/tix8.1/pref",
			files(join(sys.prefix, "tcl", "tix8.4.3", "pref"))),
		("tcl/tcl8.1/init.tcl",
			[join(sys.prefix, "tcl", "tix8.4.3", "init.tcl")]),
		("", ["about", "license-dialog", "help"]),
		("ui", glob.glob("ui/*.py")),
		("ui/glade", glob.glob("ui/glade/*")),
		("res", glob.glob("res/*")),
		("plugins", glob.glob("pyglossary/plugins/*")),
		("plugin_lib", glob.glob("pyglossary/plugin_lib/*")),
		("doc/pyglossary", ["doc/bgl_structure.svgz", ]),
		("doc/pyglossary/non-gui_examples",
			glob.glob("doc/non-gui_examples/*")),
	]
	for pyVer in ("34", "35", "36"):
		relPath = "plugin_lib/py%s" % pyVer
		data_files.append((
			relPath,
			glob.glob("pyglossary/" + relPath + "/*.py",
		)))
else:
	py2exeoptions = {}

with open("README.md", "r") as fh:
	long_description = fh.read()

setup(
	name="pyglossary",
	version=VERSION,
	cmdclass={
		"install": my_install,
	},
	description="A tool for workig with dictionary databases",
	long_description_content_type="text/markdown",
	long_description=long_description,
	author="Saeed Rasooli",
	author_email="saeed.gnu@gmail.com",
	license="GPLv3+",
	url="https://github.com/ilius/pyglossary",
	scripts=[
		# "main.py",
	],
	packages=[
		"pyglossary",
	],
	package_data={
		"pyglossary": [
			"plugins/*.py",
			"plugin_lib/*.py",
			"plugin_lib/py*/*.py",
		] + [
			# safest way found so far to include every resource of plugins
			# producing plugins/pkg/*, plugins/pkg/sub1/*, ... except .pyc/.pyo
			re.sub(
				r"^.*?pyglossary%s(?=plugins)" % ("\\\\" if os.sep == "\\" else os.sep),
				"",
				join(dirpath, f),
			)
			for top in glob.glob(
				join(dirname(__file__), "pyglossary", "plugins")
			)
			for dirpath, _, files in os.walk(top)
			for f in files
			if not (f.endswith(".pyc") or f.endswith(".pyo"))
		],
	},
	data_files=data_files,
	**py2exeoptions
)



# Copyright (c) 2019 Saeed Rasooli
# Copyright (c) 2012 Erik Rose

# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# from https://github.com/erikrose/more-itertools
def unique_everseen(iterable, key=None):
	from itertools import filterfalse
	"List unique elements, preserving order. Remember all elements ever seen."
	# unique_everseen('AAAABBBCCDAABBB') --> A B C D
	# unique_everseen('ABBCcAD', str.lower) --> A B C D
	seen = set()
	seen_add = seen.add
	if key is None:
		for element in filterfalse(seen.__contains__, iterable):
			seen_add(element)
			yield element
	else:
		for element in iterable:
			k = key(element)
			if k not in seen:
				seen_add(k)
				yield element



#pylint: disable=C0111,C0103,C0302,R0903,R0904,R0914,R0201
import argparse
import array
import collections
import encodings
import functools
import io
import os
import pickle
import random
import sys
import tempfile
import unicodedata
import unittest
import warnings

from abc import abstractmethod
from bisect import bisect_left
from builtins import open as fopen
from collections import namedtuple
from collections.abc import Sequence
from datetime import datetime, timezone
from functools import lru_cache
from struct import pack, unpack, calcsize
from threading import RLock
from types import MappingProxyType
from uuid import uuid4, UUID

import icu

DEFAULT_COMPRESSION = 'lzma2'

UTF8 = 'utf-8'
MAGIC = b'!-1SLOB\x1F'

Compression = namedtuple('Compression', 'compress decompress')

Ref = namedtuple('Ref', 'key bin_index item_index fragment')

Header = namedtuple('Header',
                    'magic uuid encoding '
                    'compression tags content_types '
                    'blob_count '
                    'store_offset '
                    'refs_offset '
                    'size')

U_CHAR = '>B'
U_CHAR_SIZE = calcsize(U_CHAR)
U_SHORT = '>H'
U_SHORT_SIZE = calcsize(U_SHORT)
U_INT = '>I'
U_INT_SIZE = calcsize(U_INT)
U_LONG_LONG = '>Q'
U_LONG_LONG_SIZE = calcsize(U_LONG_LONG)



def calcmax(len_size_spec):
    return 2**(calcsize(len_size_spec)*8) - 1

MAX_TEXT_LEN = calcmax(U_SHORT)
MAX_TINY_TEXT_LEN = calcmax(U_CHAR)
MAX_LARGE_BYTE_STRING_LEN = calcmax(U_INT)
MAX_BIN_ITEM_COUNT = calcmax(U_SHORT)

from icu import Locale, Collator, UCollAttribute, UCollAttributeValue

PRIMARY = Collator.PRIMARY
SECONDARY = Collator.SECONDARY
TERTIARY = Collator.TERTIARY
QUATERNARY = Collator.QUATERNARY
IDENTICAL = Collator.IDENTICAL


def init_compressions():
    ident = lambda x: x
    compressions = {'': Compression(ident, ident)}
    for name in ('bz2', 'zlib'):
        try:
            m = __import__(name)
        except ImportError:
            warnings.warn('%s is not available' % name)
        else:
            compressions[name] = Compression(
                lambda x: m.compress(x, 9), m.decompress)

    try:
        import lzma
    except ImportError:
        warnings.warn('lzma is not available')
    else:
        filters = [{'id': lzma.FILTER_LZMA2}]
        compress = lambda s: lzma.compress(s,
                                           format=lzma.FORMAT_RAW,
                                           filters=filters)
        decompress = lambda s: lzma.decompress(s,
                                               format=lzma.FORMAT_RAW,
                                               filters=filters)
        compressions['lzma2'] = Compression(compress, decompress)
    return compressions

COMPRESSIONS = init_compressions()


del init_compressions


MIME_TEXT = 'text/plain'
MIME_HTML = 'text/html'
MIME_CSS = 'text/css'
MIME_JS = 'application/javascript'

MIME_TYPES = {
    "html": MIME_HTML,
    "txt": MIME_TEXT,
    "js": MIME_JS,
    "css": MIME_CSS,
    "json": "application/json",
    "woff": "application/font-woff",
    "svg": "image/svg+xml",
    "png": "image/png",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "gif": "image/gif",
    "ttf": "application/x-font-ttf",
    "otf": "application/x-font-opentype"
}


class FileFormatException(Exception):
    pass


class UnknownFileFormat(FileFormatException):
    pass


class UnknownCompression(FileFormatException):
    pass


class UnknownEncoding(FileFormatException):
    pass


class IncorrectFileSize(FileFormatException):
    pass


class TagNotFound(Exception):
    pass


@lru_cache(maxsize=None)
def sortkey(strength, maxlength=None):
    c = Collator.createInstance(Locale(''))
    c.setStrength(strength)
    c.setAttribute(UCollAttribute.ALTERNATE_HANDLING,
                   UCollAttributeValue.SHIFTED)
    if maxlength is None:
        return c.getSortKey
    else:
        return lambda x: c.getSortKey(x)[:maxlength]


def sortkey_length(strength, word):
    c = Collator.createInstance(Locale(''))
    c.setStrength(strength)
    c.setAttribute(UCollAttribute.ALTERNATE_HANDLING,
                   UCollAttributeValue.SHIFTED)
    coll_key = c.getSortKey(word)
    return len(coll_key) - 1 #subtract 1 for ending \x00 byte


class MultiFileReader(io.BufferedIOBase):

    def __init__(self, *args):
        filenames = []
        for arg in args:
            if isinstance(arg, str):
                filenames.append(arg)
            else:
                for name in arg:
                    filenames.append(name)
        files = []
        ranges = []
        offset = 0
        for name in filenames:
            size = os.stat(name).st_size
            ranges.append(range(offset, offset+size))
            files.append(fopen(name, 'rb'))
            offset += size
        self.size = offset
        self._ranges = ranges
        self._files = files
        self._fcount = len(self._files)
        self._offset = -1
        self.seek(0)

    def  __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        for f in self._files:
            f.close()
        self._files.clear()
        self._ranges.clear()

    def closed(self):
        return len(self._ranges) == 0

    def isatty(self):
        return False

    def readable(self):
        return True

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            self._offset = offset
        elif whence == io.SEEK_CUR:
            self._offset = self._offset + offset
        elif whence == io.SEEK_END:
            self._offset = self.size + offset
        else:
            raise ValueError('Invalid value for parameter whence: %r' % whence)
        return self._offset

    def seekable(self):
        return True

    def tell(self):
        return self._offset

    def writable(self):
        return False

    def read(self, n=-1):
        file_index = -1
        actual_offset = 0
        for i, r in enumerate(self._ranges):
            if self._offset in r:
                file_index = i
                actual_offset = self._offset - r.start
                break
        result = b''
        if (n == -1 or n is None):
            to_read = self.size
        else:
            to_read = n
        while -1 < file_index < self._fcount:
            f = self._files[file_index]
            f.seek(actual_offset)
            read = f.read(to_read)
            read_count = len(read)
            self._offset += read_count
            result += read
            to_read -= read_count
            if to_read > 0:
                file_index += 1
                actual_offset = 0
            else:
                break
        return result


class CollationKeyList(object):

    def __init__(self, lst, sortkey_):
        self.lst = lst
        self.sortkey = sortkey_

    def __len__(self):
        return len(self.lst)

    def __getitem__(self, i):
        return self.sortkey(self.lst[i].key)


class KeydItemDict(object):

    def __init__(self, lst, strength, maxlength=None):
        self.lst = lst
        self.sortkey = sortkey(strength, maxlength=maxlength)
        self.sortkeylist = CollationKeyList(lst, self.sortkey)

    def __len__(self):
        return len(self.lst)

    def __getitem__(self, key):
        key_as_sk = self.sortkey(key)
        i = bisect_left(self.sortkeylist, key_as_sk)
        if i != len(self.lst):
            while i < len(self.lst):
                if (self.sortkey(self.lst[i].key) == key_as_sk):
                    yield self.lst[i]
                else:
                    break
                i += 1

    def __contains__(self, key):
        try:
            next(self[key])
        except StopIteration:
            return False
        else:
            return True


class Blob(object):

    def __init__(self, content_id, key, fragment,
                 read_content_type_func, read_func):
        self._content_id = content_id
        self._key = key
        self._fragment = fragment
        self._read_content_type = read_content_type_func
        self._read = read_func

    @property
    def id(self):
        return self._content_id

    @property
    def key(self):
        return self._key

    @property
    def fragment(self):
        return self._fragment

    @property
    def content_type(self):
        return self._read_content_type()

    @property
    def content(self):
        return self._read()

    def __str__(self):
        return self.key

    def __repr__(self):
        return ('<{0.__class__.__module__}.{0.__class__.__name__} '
                '{0.key}>'.format(self))


def read_byte_string(f, len_spec):
    length = unpack(len_spec, f.read(calcsize(len_spec)))[0]
    return f.read(length)


class StructReader:

    def __init__(self, file_, encoding=None):
        self._file = file_
        self.encoding = encoding

    def read_int(self):
        s = self.read(U_INT_SIZE)
        return unpack(U_INT, s)[0]

    def read_long(self):
        b = self.read(U_LONG_LONG_SIZE)
        return unpack(U_LONG_LONG, b)[0]

    def read_byte(self):
        s = self.read(U_CHAR_SIZE)
        return unpack(U_CHAR, s)[0]

    def read_short(self):
        return unpack(U_SHORT, self._file.read(U_SHORT_SIZE))[0]

    def _read_text(self, len_spec):
        max_len = 2**(8*calcsize(len_spec)) - 1
        byte_string = read_byte_string(self._file, len_spec)
        if len(byte_string) == max_len:
            terminator = byte_string.find(0)
            if terminator > -1:
                byte_string = byte_string[:terminator]
        return byte_string.decode(self.encoding)

    def read_tiny_text(self):
        return self._read_text(U_CHAR)

    def read_text(self):
        return self._read_text(U_SHORT)

    def __getattr__(self, name):
        return getattr(self._file, name)


class StructWriter:

    def __init__(self, file_, encoding=None):
        self._file = file_
        self.encoding = encoding

    def write_int(self, value):
        self._file.write(pack(U_INT, value))

    def write_long(self, value):
        self._file.write(pack(U_LONG_LONG, value))

    def write_byte(self, value):
        self._file.write(pack(U_CHAR, value))

    def write_short(self, value):
        self._file.write(pack(U_SHORT, value))

    def _write_text(self, text, len_size_spec, encoding=None, pad_to_length=None):
        if encoding is None:
            encoding = self.encoding
        text_bytes = text.encode(encoding)
        length = len(text_bytes)
        max_length = calcmax(len_size_spec)
        if length > max_length:
            raise ValueError("Text is too long for size spec %s" % len_size_spec)
        self._file.write(
            pack(len_size_spec,
                 pad_to_length if pad_to_length else length))
        self._file.write(text_bytes)
        if pad_to_length:
            for _ in range(pad_to_length - length):
                self._file.write(pack(U_CHAR, 0))

    def write_tiny_text(self, text, encoding=None, editable=False):
        pad_to_length = 255 if editable else None
        self._write_text(text, U_CHAR,
                         encoding=encoding,
                         pad_to_length=pad_to_length)

    def write_text(self, text, encoding=None):
        self._write_text(text, U_SHORT, encoding=encoding)

    def __getattr__(self, name):
        return getattr(self._file, name)


def set_tag_value(filename, name, value):
    with fopen(filename, 'rb+') as f:
        f.seek(len(MAGIC) + 16)
        encoding = read_byte_string(f, U_CHAR).decode(UTF8)
        if encodings.search_function(encoding) is None:
            raise UnknownEncoding(encoding)
        f = StructWriter(
            StructReader(f, encoding=encoding),
            encoding=encoding)
        f.read_tiny_text()
        tag_count = f.read_byte()
        for _ in range(tag_count):
            key = f.read_tiny_text()
            if key == name:
                f.write_tiny_text(value, editable=True)
                return
            f.read_tiny_text()
    raise TagNotFound(name)


def read_header(f):
    f.seek(0)

    magic = f.read(len(MAGIC))
    if (magic != MAGIC):
        raise UnknownFileFormat()

    uuid = UUID(bytes=f.read(16))
    encoding = read_byte_string(f, U_CHAR).decode(UTF8)
    if encodings.search_function(encoding) is None:
        raise UnknownEncoding(encoding)

    f = StructReader(f, encoding)
    compression = f.read_tiny_text()
    if not compression in COMPRESSIONS:
        raise UnknownCompression(compression)

    def read_tags():
        tags = {}
        count = f.read_byte()
        for _ in range(count):
            key = f.read_tiny_text()
            value = f.read_tiny_text()
            tags[key] = value
        return tags
    tags = read_tags()

    def read_content_types():
        content_types = []
        count = f.read_byte()
        for _ in range(count):
            content_type = f.read_text()
            content_types.append(content_type)
        return tuple(content_types)
    content_types = read_content_types()

    blob_count = f.read_int()
    store_offset = f.read_long()
    size = f.read_long()
    refs_offset = f.tell()

    return Header(magic=magic,
                  uuid=uuid,
                  encoding=encoding,
                  compression=compression,
                  tags=MappingProxyType(tags),
                  content_types=content_types,
                  blob_count=blob_count,
                  store_offset=store_offset,
                  refs_offset=refs_offset,
                  size=size)


def meld_ints(a, b):
    return (a << 16) | b


def unmeld_ints(c):
    bstr = bin(c).lstrip("0b").zfill(48)
    a, b = bstr[-48:-16], bstr[-16:]
    return int(a, 2), int(b, 2)


class Slob(Sequence):

    def __init__(self, file_or_filenames):
        self._f = MultiFileReader(file_or_filenames)

        try:
            self._header = read_header(self._f)
            if (self._f.size != self._header.size):
                raise IncorrectFileSize(
                    'File size should be {0}, {1} bytes found'
                    .format(self._header.size, self._f.size))
        except FileFormatException:
            self._f.close()
            raise

        self._refs = RefList(self._f,
                             self._header.encoding,
                             offset=self._header.refs_offset)

        self._g = MultiFileReader(file_or_filenames)
        self._store = Store(self._g,
                            self._header.store_offset,
                            COMPRESSIONS[self._header.compression].decompress,
                            self._header.content_types)

    def  __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @property
    def id(self):
        return self._header.uuid.hex

    @property
    def content_types(self):
        return self._header.content_types

    @property
    def tags(self):
        return self._header.tags

    @property
    def blob_count(self):
        return self._header.blob_count

    @property
    def compression(self):
        return self._header.compression

    @property
    def encoding(self):
        return self._header.encoding

    def __len__(self):
        return len(self._refs)

    def __getitem__(self, i):
        ref = self._refs[i]

        def read_func():
            return self._store.get(ref.bin_index, ref.item_index)[1]
        read_func = lru_cache(maxsize=None)(read_func)

        def read_content_type_func():
            return self._store.content_type(ref.bin_index, ref.item_index)

        content_id = meld_ints(ref.bin_index, ref.item_index)
        return Blob(content_id, ref.key, ref.fragment,
                    read_content_type_func, read_func)

    def get(self, blob_id):
        bin_index, bin_item_index = unmeld_ints(blob_id)
        return self._store.get(bin_index, bin_item_index)

    @lru_cache(maxsize=None)
    def as_dict(self,
                strength=TERTIARY,
                maxlength=None):
        return KeydItemDict(self, strength, maxlength=maxlength)

    def close(self):
        self._f.close()
        self._g.close()


def find_parts(fname):
    fname = os.path.expanduser(fname)
    dirname = os.path.dirname(fname) or os.getcwd()
    basename = os.path.basename(fname)
    candidates = []
    for name in os.listdir(dirname):
        if name.startswith(basename):
            candidates.append(os.path.join(dirname, name))
    return sorted(candidates)


def open(file_or_filenames):
    if isinstance(file_or_filenames, str):
        if not os.path.exists(file_or_filenames):
            file_or_filenames = find_parts(file_or_filenames)
    return Slob(file_or_filenames)


def create(*args, **kwargs):
    return Writer(*args, **kwargs)


class BinMemWriter:

    def __init__(self):
        self.content_type_ids = []
        self.item_dir = []
        self.items = []
        self.current_offset = 0

    def add(self, content_type_id, blob):
        self.content_type_ids.append(content_type_id)
        self.item_dir.append(pack(U_INT, self.current_offset))
        length_and_bytes = pack(U_INT, len(blob)) + blob
        self.items.append(length_and_bytes)
        self.current_offset += len(length_and_bytes)

    def __len__(self):
        return len(self.item_dir)

    def finalize(self, fout: 'output file', compress: 'function'):
        count = len(self)
        fout.write(pack(U_INT, count))
        for content_type_id in self.content_type_ids:
            fout.write(pack(U_CHAR, content_type_id))
        content = b''.join(self.item_dir + self.items)
        compressed = compress(content)
        fout.write(pack(U_INT, len(compressed)))
        fout.write(compressed)
        self.content_type_ids.clear()
        self.item_dir.clear()
        self.items.clear()


class ItemList(Sequence):

    def __init__(self, file_, offset,
                 count_or_spec, pos_spec,
                 cache_size=None):
        self.lock = RLock()
        self._file = file_
        file_.seek(offset)
        if isinstance(count_or_spec, str):
            count_spec = count_or_spec
            self.count = unpack(count_spec, file_.read(calcsize(count_spec)))[0]
        else:
            self.count = count_or_spec
        self.pos_offset = file_.tell()
        self.pos_spec = pos_spec
        self.pos_size = calcsize(pos_spec)
        self.data_offset = self.pos_offset + self.pos_size * self.count
        if cache_size:
            self.__getitem__ = lru_cache(maxsize=cache_size)(self.__getitem__)

    def __len__(self):
        return self.count

    def pos(self, i):
        with self.lock:
            self._file.seek(self.pos_offset + self.pos_size*i)
            return unpack(self.pos_spec, self._file.read(self.pos_size))[0]

    def read(self, pos):
        with self.lock:
            self._file.seek(self.data_offset + pos)
            return self._read_item()

    @abstractmethod
    def _read_item(self):
        pass

    def __getitem__(self, i):
        if i >= len(self) or i < 0:
            raise IndexError('index out of range')
        return self.read(self.pos(i))


class RefList(ItemList):

    def __init__(self, f, encoding, offset=0, count=None):
        super().__init__(StructReader(f, encoding),
                         offset,
                         U_INT if count is None else count,
                         U_LONG_LONG,
                         cache_size=512)


    def _read_item(self):
        key = self._file.read_text()
        bin_index = self._file.read_int()
        item_index = self._file.read_short()
        fragment = self._file.read_tiny_text()
        return Ref(key=key,
                   bin_index=bin_index,
                   item_index=item_index,
                   fragment=fragment)

    @lru_cache(maxsize=None)
    def as_dict(self,
                strength=TERTIARY,
                maxlength=None):
        return KeydItemDict(self, strength, maxlength=maxlength)


class Bin(ItemList):

    def __init__(self, count, bin_bytes):
        super().__init__(StructReader(io.BytesIO(bin_bytes)),
                         0,
                         count,
                         U_INT)

    def _read_item(self):
        content_len = self._file.read_int()
        content = self._file.read(content_len)
        return content


StoreItem = namedtuple('StoreItem', 'content_type_ids compressed_content')


class Store(ItemList):

    def __init__(self, file_, offset, decompress, content_types):
        super().__init__(StructReader(file_),
                         offset,
                         U_INT,
                         U_LONG_LONG,
                         cache_size=32)
        self.decompress = decompress
        self.content_types = content_types

    def _read_item(self):
        bin_item_count = self._file.read_int()
        packed_content_type_ids = self._file.read(bin_item_count*U_CHAR_SIZE)
        content_type_ids = []
        for i in range(bin_item_count):
            content_type_id = unpack(U_CHAR, packed_content_type_ids[i:i+1])[0]
            content_type_ids.append(content_type_id)
        content_length = self._file.read_int()
        content = self._file.read(content_length)
        return StoreItem(content_type_ids=content_type_ids,
                         compressed_content=content)

    def _content_type(self, bin_index, item_index):
        store_item = self[bin_index]
        content_type_id = store_item.content_type_ids[item_index]
        content_type = self.content_types[content_type_id]
        return content_type, store_item

    def content_type(self, bin_index, item_index):
        return self._content_type(bin_index, item_index)[0]

    @lru_cache(maxsize=16)
    def _decompress(self, bin_index):
        store_item = self[bin_index]
        return self.decompress(store_item.compressed_content)

    def get(self, bin_index, item_index):
        content_type, store_item = self._content_type(bin_index, item_index)
        content = self._decompress(bin_index)
        count = len(store_item.content_type_ids)
        store_bin = Bin(count, content)
        content = store_bin[item_index]
        return (content_type, content)


def find(word, slobs, match_prefix=True):
    seen = set()
    if isinstance(slobs, Slob):
        slobs = [slobs]

    variants = []

    for strength in (QUATERNARY, TERTIARY, SECONDARY, PRIMARY):
        variants.append((strength, None))

    if match_prefix:
        for strength in (QUATERNARY, TERTIARY, SECONDARY, PRIMARY):
            variants.append((strength, sortkey_length(strength, word)))

    for strength, maxlength in variants:
        for slob in slobs:
            d = slob.as_dict(strength=strength, maxlength=maxlength)
            for item in d[word]:
                dedup_key = (slob.id, item.id, item.fragment)
                if dedup_key in seen:
                    continue
                else:
                    seen.add(dedup_key)
                    yield slob, item


WriterEvent = namedtuple('WriterEvent', 'name data')


class KeyTooLongException(Exception):

    @property
    def key(self):
        return self.args[0]


class Writer(object):

    def __init__(self,
                 filename,
                 workdir=None,
                 encoding=UTF8,
                 compression=DEFAULT_COMPRESSION,
                 min_bin_size=512*1024,
                 max_redirects=5,
                 observer=None):
        self.filename = filename
        self.observer = observer
        if os.path.exists(self.filename):
            raise SystemExit('File %r already exists' % self.filename)

        #make sure we can write
        with fopen(self.filename, 'wb'):
            pass

        self.encoding = encoding

        if encodings.search_function(self.encoding) is None:
            raise UnknownEncoding(self.encoding)

        self.workdir = workdir

        self.tmpdir = tmpdir = tempfile.TemporaryDirectory(
            prefix='{0}-'.format(os.path.basename(filename)),
            dir=workdir)

        self.f_ref_positions = self._wbfopen('ref-positions')
        self.f_store_positions = self._wbfopen('store-positions')
        self.f_refs = self._wbfopen('refs')
        self.f_store = self._wbfopen('store')

        self.max_redirects = max_redirects
        if max_redirects:
            self.aliases_path = os.path.join(tmpdir.name, 'aliases')
            self.f_aliases = Writer(self.aliases_path,
                                    workdir=tmpdir.name,
                                    max_redirects=0,
                                    compression=None)

        if compression is None:
            compression = ''
        if not compression in COMPRESSIONS:
            raise UnknownCompression(compression)
        else:
            self.compress = COMPRESSIONS[compression].compress

        self.compression = compression
        self.content_types = {}

        self.min_bin_size = min_bin_size

        self.current_bin = None

        self.blob_count = 0
        self.ref_count = 0
        self.bin_count = 0
        self._tags = {
            'version.python': sys.version.replace('\n', ' '),
            'version.pyicu': icu.VERSION,
            'version.icu': icu.ICU_VERSION,
            'created.at': datetime.now(timezone.utc).isoformat()
        }
        self.tags = MappingProxyType(self._tags)

    def _wbfopen(self, name):
        return StructWriter(
            fopen(os.path.join(self.tmpdir.name, name), 'wb'),
            encoding=self.encoding)

    def tag(self, name, value=''):
        if len(name.encode(self.encoding)) > MAX_TINY_TEXT_LEN:
            self._fire_event('tag_name_too_long', (name, value))
            return

        if len(value.encode(self.encoding)) > MAX_TINY_TEXT_LEN:
            self._fire_event('tag_value_too_long', (name, value))
            value = ''

        self._tags[name] = value

    def _split_key(self, key):
        if isinstance(key, str):
            actual_key = key
            fragment = ''
        else:
            actual_key, fragment = key
        if (len(actual_key) > MAX_TEXT_LEN or
            len(fragment) > MAX_TINY_TEXT_LEN):
            raise KeyTooLongException(key)
        return actual_key, fragment

    def add(self, blob, *keys, content_type=''):

        if len(blob) > MAX_LARGE_BYTE_STRING_LEN:
            self._fire_event('content_too_long', blob)
            return

        if len(content_type) > MAX_TEXT_LEN:
            self._fire_event('content_type_too_long', content_type)
            return

        actual_keys = []

        for key in keys:
            try:
                actual_key, fragment = self._split_key(key)
            except KeyTooLongException as e:
                self._fire_event('key_too_long', e.key)
            else:
                actual_keys.append((actual_key, fragment))

        if len(actual_keys) == 0:
            return

        if self.current_bin is None:
            self.current_bin = BinMemWriter()
            self.bin_count += 1

        if content_type not in self.content_types:
            self.content_types[content_type] = len(self.content_types)

        self.current_bin.add(self.content_types[content_type], blob)
        self.blob_count += 1
        bin_item_index = len(self.current_bin) - 1
        bin_index = self.bin_count - 1

        for actual_key, fragment in actual_keys:
            self._write_ref(actual_key, bin_index, bin_item_index, fragment)

        if (self.current_bin.current_offset > self.min_bin_size or
            len(self.current_bin) == MAX_BIN_ITEM_COUNT):
            self._write_current_bin()

    def add_alias(self, key, target_key):
        if self.max_redirects:
            try:
                self._split_key(key)
            except KeyTooLongException as e:
                self._fire_event('alias_too_long', e.key)
                return
            try:
                self._split_key(target_key)
            except KeyTooLongException as e:
                self._fire_event('alias_target_too_long', e.key)
                return
            self.f_aliases.add(pickle.dumps(target_key), key)
        else:
            raise NotImplementedError()

    def _fire_event(self, name, data=None):
        if self.observer:
            self.observer(WriterEvent(name, data))

    def _write_current_bin(self):
        self.f_store_positions.write_long(self.f_store.tell())
        self.current_bin.finalize(self.f_store, self.compress)
        self.current_bin = None

    def _write_ref(self, key, bin_index, item_index, fragment=''):
        self.f_ref_positions.write_long(self.f_refs.tell())
        self.f_refs.write_text(key)
        self.f_refs.write_int(bin_index)
        self.f_refs.write_short(item_index)
        self.f_refs.write_tiny_text(fragment)
        self.ref_count += 1

    def _sort(self):
        self._fire_event('begin_sort')
        f_ref_positions_sorted = self._wbfopen('ref-positions-sorted')
        self.f_refs.flush()
        self.f_ref_positions.close()
        with MultiFileReader(self.f_ref_positions.name, self.f_refs.name) as f:
            ref_list = RefList(f, self.encoding, count=self.ref_count)
            sortkey_func = sortkey(IDENTICAL)
            for i in sorted(range(len(ref_list)),
                            key=lambda j: sortkey_func(ref_list[j].key)):
                ref_pos = ref_list.pos(i)
                f_ref_positions_sorted.write_long(ref_pos)
        f_ref_positions_sorted.close()
        os.remove(self.f_ref_positions.name)
        os.rename(f_ref_positions_sorted.name, self.f_ref_positions.name)
        self.f_ref_positions = StructWriter(
            fopen(self.f_ref_positions.name, 'ab'),
            encoding=self.encoding)
        self._fire_event('end_sort')

    def _resolve_aliases(self):
        self._fire_event('begin_resolve_aliases')
        self.f_aliases.finalize()
        with MultiFileReader(self.f_ref_positions.name,
                             self.f_refs.name) as f_ref_list:
            ref_list = RefList(f_ref_list, self.encoding, count=self.ref_count)
            ref_dict = ref_list.as_dict()
            with open(self.aliases_path) as r:
                aliases = r.as_dict()
                path = os.path.join(self.tmpdir.name, 'resolved-aliases')
                with create(path,
                            workdir=self.tmpdir.name,
                            max_redirects=0,
                            compression=None) as alias_writer:

                    def read_key_frag(item, default_fragment):
                        key_frag = pickle.loads(item.content)
                        if isinstance(key_frag, str):
                            return key_frag, default_fragment
                        else:
                            return key_frag

                    for item in r:
                        from_key = item.key
                        keys = set()
                        keys.add(from_key)
                        to_key, fragment = read_key_frag(item, item.fragment)
                        count = 0
                        while count <= self.max_redirects:
                            #is target key itself a redirect?
                            try:
                                orig_to_key = to_key
                                to_key, fragment = read_key_frag(
                                    next(aliases[to_key]), fragment)
                                count += 1
                                keys.add(orig_to_key)
                            except StopIteration:
                                break
                        if count > self.max_redirects:
                            self._fire_event('too_many_redirects', from_key)
                        try:
                            target_ref = next(ref_dict[to_key])
                        except StopIteration:
                            self._fire_event('alias_target_not_found', to_key)
                        else:
                            for key in keys:
                                ref = Ref(key=key,
                                          bin_index=target_ref.bin_index,
                                          item_index=target_ref.item_index,
                                          #last fragment in the chain wins
                                          fragment=target_ref.fragment or fragment)
                                alias_writer.add(pickle.dumps(ref), key)

        with open(path) as resolved_aliases_reader:
            previous_key = None
            for item in resolved_aliases_reader:
                ref = pickle.loads(item.content)
                if ref.key == previous_key:
                    continue
                self._write_ref(ref.key, ref.bin_index,
                                ref.item_index, ref.fragment)
                previous_key = ref.key
        self._sort()
        self._fire_event('end_resolve_aliases')


    def finalize(self):
        self._fire_event('begin_finalize')
        if not self.current_bin is None:
            self._write_current_bin()

        self._sort()
        if self.max_redirects:
            self._resolve_aliases()

        files = (self.f_ref_positions,
                 self.f_refs,
                 self.f_store_positions,
                 self.f_store)

        for f in files:
            f.close()

        buf_size = 10*1024*1024

        with fopen(self.filename, mode='wb') as output_file:
            out = StructWriter(output_file, self.encoding)
            out.write(MAGIC)
            out.write(uuid4().bytes)
            out.write_tiny_text(self.encoding, encoding=UTF8)
            out.write_tiny_text(self.compression)

            def write_tags(tags, f):
                f.write(pack(U_CHAR, len(tags)))
                for key, value in tags.items():
                    f.write_tiny_text(key)
                    f.write_tiny_text(value, editable=True)
            write_tags(self.tags, out)

            def write_content_types(content_types, f):
                count = len(content_types)
                f.write(pack(U_CHAR, count))
                types = sorted(content_types.items(), key=lambda x: x[1])
                for content_type, _ in types:
                    f.write_text(content_type)
            write_content_types(self.content_types, out)

            out.write_int(self.blob_count)
            store_offset = (out.tell() +
                                    U_LONG_LONG_SIZE + # this value
                                    U_LONG_LONG_SIZE + # file size value
                                    U_INT_SIZE + # ref count value
                                    os.stat(self.f_ref_positions.name).st_size +
                                    os.stat(self.f_refs.name).st_size)
            out.write_long(store_offset)
            out.flush()

            file_size = (out.tell() + # bytes written so far
                         U_LONG_LONG_SIZE + # file size value
                         2*U_INT_SIZE) # ref count and bin count
            file_size += sum((os.stat(f.name).st_size for f in files))
            out.write_long(file_size)

            def mv(src, out):
                fname = src.name
                self._fire_event('begin_move', fname)
                with fopen(fname, mode='rb') as f:
                    while True:
                        data = f.read(buf_size)
                        if len(data) == 0:
                            break
                        out.write(data)
                        out.flush()
                os.remove(fname)
                self._fire_event('end_move', fname)

            out.write_int(self.ref_count)
            mv(self.f_ref_positions, out)
            mv(self.f_refs, out)

            out.write_int(self.bin_count)
            mv(self.f_store_positions, out)
            mv(self.f_store, out)


        self.tmpdir.cleanup()
        self._fire_event('end_finalize')


    def size_header(self):
        size = 0
        size += len(MAGIC)
        size += 16 # uuid bytes
        size += U_CHAR_SIZE + len(self.encoding.encode(UTF8))
        size += U_CHAR_SIZE + len(self.compression.encode(self.encoding))

        size += U_CHAR_SIZE # tag length
        size += U_CHAR_SIZE # content types count

        #tags and content types themselves counted elsewhere

        size += U_INT_SIZE #blob count
        size += U_LONG_LONG_SIZE #store offset
        size += U_LONG_LONG_SIZE # file size
        size += U_INT_SIZE # ref count
        size += U_INT_SIZE # bin count

        return size

    def size_tags(self):
        size = 0
        for key, _ in self.tags.items():
            size += U_CHAR_SIZE + len(key.encode(self.encoding))
            size += 255
        return size

    def size_content_types(self):
        size = 0
        for content_type in self.content_types:
            size += U_CHAR_SIZE + len(content_type.encode(self.encoding))
        return size

    def size_data(self):
        files = (self.f_ref_positions,
                 self.f_refs,
                 self.f_store_positions,
                 self.f_store)
        return sum((os.stat(f.name).st_size for f in files))

    def  __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finalize()
        return False


class TestReadWrite(unittest.TestCase):

    def setUp(self):

        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')

        with create(self.path) as w:

            self.tags = {'a': 'abc',
                         'bb': 'xyz123',
                         'ccc': 'lkjlk'}
            for name, value in self.tags.items():
                w.tag(name, value)

            self.tag2 = 'bb', 'xyz123'

            self.blob_encoding = 'ascii'

            self.data = [
                (('c', 'cc', 'ccc'), MIME_TEXT, 'Hello C 1'),
                ('a', MIME_TEXT, 'Hello A 12'),
                ('z', MIME_TEXT, 'Hello Z 123'),
                ('b', MIME_TEXT, 'Hello B 1234'),
                ('d', MIME_TEXT, 'Hello D 12345'),
                ('uuu', MIME_HTML, '<html><body>Hello U!</body></html>'),
                ((('yy', 'frag1'),), MIME_HTML, '<h1 name="frag1">Section 1</h1>'),
            ]

            self.all_keys = []

            self.data_as_dict = {}

            for k, t, v in self.data:
                if isinstance(k, str):
                    k = (k,)
                for key in k:
                    if isinstance(key, tuple):
                        key, fragment = key
                    else:
                        fragment = ''
                    self.all_keys.append(key)
                    self.data_as_dict[key] = (t, v, fragment)
                w.add(v.encode(self.blob_encoding), *k, content_type=t)
            self.all_keys.sort()

        self.w = w

    def test_header(self):
        with MultiFileReader(self.path) as f:
            header = read_header(f)

        for key, value in self.tags.items():
            self.assertEqual(header.tags[key], value)

        self.assertEqual(self.w.encoding, UTF8)
        self.assertEqual(header.encoding, self.w.encoding)

        self.assertEqual(header.compression, self.w.compression)

        for i, content_type in enumerate(header.content_types):
            self.assertEqual(self.w.content_types[content_type], i)

        self.assertEqual(header.blob_count, len(self.data))

    def test_content(self):
        with open(self.path) as r:
            self.assertEqual(len(r), len(self.all_keys))
            self.assertRaises(IndexError,
                              r.__getitem__, len(self.all_keys))
            for i, item in enumerate(r):
                self.assertEqual(item.key, self.all_keys[i])
                content_type, value, fragment = self.data_as_dict[item.key]
                self.assertEqual(
                    item.content_type, content_type)
                self.assertEqual(
                    item.content.decode(self.blob_encoding), value)
                self.assertEqual(
                    item.fragment, fragment)


    def tearDown(self):
        self.tmpdir.cleanup()


class TestSort(unittest.TestCase):

    def setUp(self):

        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')

        with create(self.path) as  w:

            data = [
                'Ф, ф',
                'Ф ф',
                'Ф',
                'Э',
                'Е е',
                'г',
                'н',
                'ф',
                'а',
                'Ф, Ф',
                'е',
                'Е',
                'Ее',
                'ё',
                'Ё',
                'Её',
                'Е ё',
                'А',
                'э',
                'ы'
            ]

            self.data_sorted = sorted(data, key=sortkey(IDENTICAL))

            for k in data:
                v = ';'.join(unicodedata.name(c) for c in k)
                w.add(v.encode('ascii'), k)

        self.r = open(self.path)

    def test_sort_order(self):
        for i in range(len(self.r)):
            self.assertEqual(self.r[i].key, self.data_sorted[i])

    def tearDown(self):
        self.r.close()
        self.tmpdir.cleanup()


class TestFind(unittest.TestCase):

    def setUp(self):

        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')

        with create(self.path) as w:
            data = ['Cc', 'aA', 'aa', 'Aa', 'Bb', 'cc', 'Äā', 'ăÀ',
                    'a\u00A0a', 'a-a', 'a\u2019a', 'a\u2032a', 'a,a', 'a a']

            for k in data:
                v = ';'.join(unicodedata.name(c) for c in k)
                w.add(v.encode('ascii'), k)

        self.r = open(self.path)

    def get(self, d, key):
        return list(item.content.decode('ascii') for item in d[key])

    def test_find_identical(self):
        d = self.r.as_dict(IDENTICAL)
        self.assertEqual(
            self.get(d, 'aa'),
            ['LATIN SMALL LETTER A;LATIN SMALL LETTER A'])
        self.assertEqual(
            self.get(d, 'a-a'),
            ['LATIN SMALL LETTER A;HYPHEN-MINUS;LATIN SMALL LETTER A'])
        self.assertEqual(
            self.get(d, 'aA'),
            ['LATIN SMALL LETTER A;LATIN CAPITAL LETTER A'])
        self.assertEqual(
            self.get(d, 'Äā'),
            ['LATIN CAPITAL LETTER A WITH DIAERESIS;'
             'LATIN SMALL LETTER A WITH MACRON'])
        self.assertEqual(
            self.get(d, 'a a'),
            ['LATIN SMALL LETTER A;SPACE;LATIN SMALL LETTER A'])

    def test_find_quaternary(self):
        d = self.r.as_dict(QUATERNARY)
        self.assertEqual(
            self.get(d, 'a\u2032a'),
            ['LATIN SMALL LETTER A;PRIME;LATIN SMALL LETTER A'])
        self.assertEqual(
            self.get(d, 'a a'),
            ['LATIN SMALL LETTER A;SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;NO-BREAK SPACE;LATIN SMALL LETTER A'])

    def test_find_tertiary(self):
        d = self.r.as_dict(TERTIARY)
        self.assertEqual(
            self.get(d, 'aa'),
            ['LATIN SMALL LETTER A;SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;NO-BREAK SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;HYPHEN-MINUS;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;COMMA;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;RIGHT SINGLE QUOTATION MARK;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;PRIME;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;LATIN SMALL LETTER A'])

    def test_find_secondary(self):
        d = self.r.as_dict(SECONDARY)
        self.assertEqual(
            self.get(d, 'aa'),
            ['LATIN SMALL LETTER A;SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;NO-BREAK SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;HYPHEN-MINUS;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;COMMA;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;RIGHT SINGLE QUOTATION MARK;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;PRIME;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;LATIN CAPITAL LETTER A',
             'LATIN CAPITAL LETTER A;LATIN SMALL LETTER A'])


    def test_find_primary(self):
        d = self.r.as_dict(PRIMARY)

        self.assertEqual(
            self.get(d, 'aa'),
            ['LATIN SMALL LETTER A;SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;NO-BREAK SPACE;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;HYPHEN-MINUS;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;COMMA;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;RIGHT SINGLE QUOTATION MARK;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;PRIME;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A;LATIN CAPITAL LETTER A',
             'LATIN CAPITAL LETTER A;LATIN SMALL LETTER A',
             'LATIN SMALL LETTER A WITH BREVE;LATIN CAPITAL LETTER A WITH GRAVE',
             'LATIN CAPITAL LETTER A WITH DIAERESIS;LATIN SMALL LETTER A WITH MACRON'])

    def tearDown(self):
        self.r.close()
        self.tmpdir.cleanup()


class TestPrefixFind(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')
        self.data = ['a', 'ab', 'abc', 'abcd', 'abcde']
        with create(self.path) as w:
            for k in self.data:
                w.add(k.encode('ascii'), k)

    def tearDown(self):
        self.tmpdir.cleanup()

    def test(self):
        with open(self.path) as r:
            for i, k in enumerate(self.data):
                d = r.as_dict(IDENTICAL, len(k))
                self.assertEqual(list(v.content.decode('ascii') for v in  d[k]),
                                 self.data[i:])


class TestBestMatch(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path1 = os.path.join(self.tmpdir.name, 'test1.slob')
        self.path2 = os.path.join(self.tmpdir.name, 'test2.slob')

        data1 = ['aa', 'Aa', 'a-a', 'aabc', 'Äā', 'bb', 'aa']
        data2 = ['aa', 'aA', 'āā', 'a,a', 'a-a', 'aade', 'Äā', 'cc']

        with create(self.path1) as w:
            for key in data1:
                w.add(b'', key)

        with create(self.path2) as w:
            for key in data2:
                w.add(b'', key)

    def test_best_match(self):
        self.maxDiff = None
        with open(self.path1) as s1, \
             open(self.path2) as s2:
            result = find('aa', [s1, s2], match_prefix=True)
            actual = list((s.id, item.key) for s, item in result)
            expected = [(s1.id, 'aa'),
                        (s1.id, 'aa'),
                        (s2.id, 'aa'),
                        (s1.id, 'a-a'),
                        (s2.id, 'a-a'),
                        (s2.id, 'a,a'),
                        (s1.id, 'Aa'),
                        (s2.id, 'aA'),
                        (s1.id, 'Äā'),
                        (s2.id, 'Äā'),
                        (s2.id, 'āā'),
                        (s1.id, 'aabc'),
                        (s2.id, 'aade'),]
            self.assertEqual(expected, actual)

    def tearDown(self):
        self.tmpdir.cleanup()


class TestAlias(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_alias(self):
        too_many_redirects = []
        target_not_found = []

        def observer(event):
            if event.name == 'too_many_redirects':
                too_many_redirects.append(event.data)
            elif event.name == 'alias_target_not_found':
                target_not_found.append(event.data)

        with create(self.path, observer=observer) as w:
            data = ['z', 'b', 'q', 'a', 'u', 'g', 'p', 'n']
            for k in data:
                v = ';'.join(unicodedata.name(c) for c in k)
                w.add(v.encode('ascii'), k)

            w.add_alias('w', 'u')
            w.add_alias('y1', 'y2')
            w.add_alias('y2', 'y3')
            w.add_alias('y3', 'z')
            w.add_alias('ZZZ', 'YYY')

            w.add_alias('l3', 'l1')
            w.add_alias('l1', 'l2')
            w.add_alias('l2', 'l3')

            w.add_alias('a1', ('a', 'a-frag1'))
            w.add_alias('a2', 'a1')
            w.add_alias('a3', ('a2', 'a-frag2'))

            w.add_alias('g1', 'g')
            w.add_alias('g2', ('g1', 'g-frag1'))


        self.assertEqual(too_many_redirects, ['l1', 'l2', 'l3'])
        self.assertEqual(target_not_found, ['l2', 'l3', 'l1', 'YYY'])

        with open(self.path) as r:
            d = r.as_dict()
            def get(key):
                return list(item.content.decode('ascii') for item in d[key])
            self.assertEqual(get('w'), ['LATIN SMALL LETTER U'])
            self.assertEqual(get('y1'), ['LATIN SMALL LETTER Z'])
            self.assertEqual(get('y2'), ['LATIN SMALL LETTER Z'])
            self.assertEqual(get('y3'), ['LATIN SMALL LETTER Z'])
            self.assertEqual(get('ZZZ'), [])
            self.assertEqual(get('l1'), [])
            self.assertEqual(get('l2'), [])
            self.assertEqual(get('l3'), [])

            item_a1 = next(d['a1'])
            self.assertEqual(item_a1.content, b'LATIN SMALL LETTER A')
            self.assertEqual(item_a1.fragment, 'a-frag1')

            item_a2 = next(d['a2'])
            self.assertEqual(item_a2.content, b'LATIN SMALL LETTER A')
            self.assertEqual(item_a2.fragment, 'a-frag1')

            item_a3 = next(d['a3'])
            self.assertEqual(item_a3.content, b'LATIN SMALL LETTER A')
            self.assertEqual(item_a3.fragment, 'a-frag1')

            item_g1 = next(d['g1'])
            self.assertEqual(item_g1.content, b'LATIN SMALL LETTER G')
            self.assertEqual(item_g1.fragment, '')

            item_g2 = next(d['g2'])
            self.assertEqual(item_g2.content, b'LATIN SMALL LETTER G')
            self.assertEqual(item_g2.fragment, 'g-frag1')


class TestBlobId(unittest.TestCase):

    def test(self):
        max_i = 2**32 - 1
        max_j = 2**16 - 1
        i_values = [0, max_i] + [random.randint(1, max_i - 1)
                                 for _ in range(100)]
        j_values = [0, max_j] + [random.randint(1, max_j - 1)
                                 for _ in range(100)]
        for i in i_values:
            for j in j_values:
                self.assertEqual(unmeld_ints(meld_ints(i, j)), (i, j))


class TestMultiFileReader(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_read_all(self):
        fnames = []
        for name in 'abcdef':
            path = os.path.join(self.tmpdir.name, name)
            fnames.append(path)
            with fopen(path, 'wb') as f:
                f.write(name.encode(UTF8))
        with MultiFileReader(fnames) as m:
            self.assertEqual(m.read().decode(UTF8), 'abcdef')

    def test_seek_and_read(self):

        def mkfile(basename, content):
            part = os.path.join(self.tmpdir.name, basename)
            with fopen(part, 'wb') as f:
                f.write(content)
            return part

        content = b'abc\nd\nefgh\nij'
        part1 = mkfile('1', content[:4])
        part2 = mkfile('2', content[4:5])
        part3 = mkfile('3', content[5:])

        with MultiFileReader(part1, part2, part3) as m:
            self.assertEqual(m.size, len(content))
            m.seek(2)
            self.assertEqual(m.read(2), content[2:4])
            m.seek(1)
            self.assertEqual(m.read(len(content) - 2), content[1:-1])
            m.seek(-1, whence=io.SEEK_END)
            self.assertEqual(m.read(10), content[-1:])

            m.seek(4)
            m.seek(-2, whence=io.SEEK_CUR)
            self.assertEqual(m.read(3), content[2:5])


class TestFormatErrors(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_wrong_file_type(self):
        name = os.path.join(self.tmpdir.name, '1')
        with fopen(name, 'wb') as f:
            f.write(b'123')
        self.assertRaises(UnknownFileFormat, open, name)

    def test_truncated_file(self):
        name = os.path.join(self.tmpdir.name, '1')

        with create(name) as f:
            f.add(b'123', 'a')
            f.add(b'234', 'b',)

        with fopen(name, 'rb') as f:
            all_bytes = f.read()

        with fopen(name, 'wb') as f:
            f.write(all_bytes[:-1])

        self.assertRaises(IncorrectFileSize, open, name)

        with fopen(name, 'wb') as f:
            f.write(all_bytes)
            f.write(b'\n')

        self.assertRaises(IncorrectFileSize, open, name)


class TestFindParts(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_find_parts(self):
        names = [os.path.join(self.tmpdir.name, name)
                 for name in ('abc-1', 'abc-2', 'abc-3')]
        for name in names:
            with fopen(name, 'wb'):
                pass
        parts = find_parts(os.path.join(self.tmpdir.name, 'abc'))
        self.assertEqual(names, parts)


class TestTooLongText(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_too_long(self):
        rejected_keys = []
        rejected_aliases = []
        rejected_alias_targets = []
        rejected_tags = []
        rejected_content_types = []
        def observer(event):
            if event.name == 'key_too_long':
                rejected_keys.append(event.data)
            elif event.name == 'alias_too_long':
                rejected_aliases.append(event.data)
            elif event.name == 'alias_target_too_long':
                rejected_alias_targets.append(event.data)
            elif event.name == 'tag_name_too_long':
                rejected_tags.append(event.data)
            elif event.name == 'content_type_too_long':
                rejected_content_types.append(event.data)

        long_tag_name = 't'*(MAX_TINY_TEXT_LEN+1)
        long_tag_value = 'v'*(MAX_TINY_TEXT_LEN+1)
        long_content_type = 'T'*(MAX_TEXT_LEN+1)
        long_key = 'c'*(MAX_TEXT_LEN+1)
        long_frag = 'd'*(MAX_TINY_TEXT_LEN+1)
        key_with_long_frag = ('d', long_frag)
        tag_with_long_name = (long_tag_name, 't3 value')
        tag_with_long_value = ('t1', long_tag_value)
        long_alias = 'f'*(MAX_TEXT_LEN+1)
        alias_with_long_frag = ('i', long_frag)
        long_alias_target = long_key
        long_alias_target_frag = key_with_long_frag

        with create(self.path, observer=observer) as w:

            w.tag(*tag_with_long_value)
            w.tag('t2', 't2 value')
            w.tag(*tag_with_long_name)

            data = ['a', 'b', long_key, key_with_long_frag]

            for k in data:
                if isinstance(k, str):
                    v = k.encode('ascii')
                else:
                    v = '#'.join(k).encode('ascii')
                w.add(v, k)

            w.add_alias('e', 'a')
            w.add_alias(long_alias, 'a')
            w.add_alias(alias_with_long_frag, 'a')
            w.add_alias('g', long_alias_target)
            w.add_alias('h', long_alias_target_frag)

            w.add(b'Hello', 'hello', content_type=long_content_type)

        self.assertEqual(rejected_keys,
                         [long_key, key_with_long_frag])
        self.assertEqual(rejected_aliases,
                         [long_alias, alias_with_long_frag])
        self.assertEqual(rejected_alias_targets,
                         [long_alias_target, long_alias_target_frag])
        self.assertEqual(rejected_tags,
                         [tag_with_long_name])
        self.assertEqual(rejected_content_types,
                         [long_content_type])

        with open(self.path) as r:
            self.assertEqual(r.tags['t2'], 't2 value')
            self.assertFalse(tag_with_long_name[0] in r.tags)
            self.assertTrue(tag_with_long_value[0] in r.tags)
            self.assertEqual(r.tags[tag_with_long_value[0]], '')
            d = r.as_dict()
            self.assertTrue('a' in d)
            self.assertTrue('b' in d)
            self.assertFalse(long_key in d)
            self.assertFalse(key_with_long_frag[0] in d)
            self.assertTrue('e' in d)
            self.assertFalse(long_alias in d)
            self.assertFalse('g' in d)

        self.assertRaises(ValueError, set_tag_value, self.path, 't1', 'ы'*128)


class TestEditTag(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')
        with create(self.path) as w:
            w.tag('a', '123456')
            w.tag('b', '654321')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_edit_existing_tag(self):
        with open(self.path) as f:
            self.assertEqual(f.tags['a'], '123456')
            self.assertEqual(f.tags['b'], '654321')
        set_tag_value(self.path, 'b', 'efg')
        set_tag_value(self.path, 'a', 'xyz')
        with open(self.path) as f:
            self.assertEqual(f.tags['a'], 'xyz')
            self.assertEqual(f.tags['b'], 'efg')

    def test_edit_nonexisting_tag(self):
        self.assertRaises(TagNotFound, set_tag_value, self.path, 'z', 'abc')


class TestBinItemNumberLimit(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix='test')
        self.path = os.path.join(self.tmpdir.name, 'test.slob')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_writing_more_then_max_number_of_bin_items(self):
        with create(self.path) as w:
            for _ in range(MAX_BIN_ITEM_COUNT+2):
                w.add(b'a', 'a')
            self.assertEqual(w.bin_count, 2)


def _cli_info(args):
    from collections import OrderedDict
    with open(args.path) as s:
        h = s._header
        print('\n')
        info = OrderedDict(
            (('id', s.id),
             ('encoding', h.encoding),
             ('compression', h.compression),
             ('blob count', s.blob_count),
             ('ref count', len(s))))
        _print_title(args.path)
        _print_dict(info)
        print('\n')
        _print_title('CONTENT TYPES')
        for ct in h.content_types:
            print(ct)
        print('\n')
        _print_title('TAGS')
        _print_tags(s)
        print('\n')

def _print_title(title):
    print(title)
    print('-'*len(title))

def _cli_tag(args):
    tag_name = args.name
    if args.value:
        try:
            set_tag_value(args.filename, tag_name, args.value)
        except TagNotFound:
            print('No such tag')
    else:
        with open(args.filename) as s:
            _print_tags(s, tag_name)


def _print_tags(s, tag_name=None):
    if tag_name:
        try:
            value = s.tags[tag_name]
        except KeyError:
            print('No such tag')
        else:
            print(value)
    else:
        _print_dict(s.tags)


def _print_dict(d):
    max_key_len = 0
    for k, v in d.items():
        key_len = len(k)
        if key_len > max_key_len:
            max_key_len = key_len
    fmt_template = '{:>%s}: {}'
    fmt = fmt_template % max_key_len
    for k, v in d.items():
        print(fmt.format(k, v))


def _cli_find(args):
    with open(args.path) as s:
        match_prefix = not args.whole
        for i, item in enumerate(find(args.query, s,
                                      match_prefix=match_prefix)):
            _, blob = item
            print(blob.id, blob.content_type, blob.key)
            if i == args.limit:
                break

def _cli_aliases(args):
    word = args.query
    with open(args.path) as s:
        d = s.as_dict(strength=QUATERNARY)
        length = len(s)
        aliases = []

        def print_item(i, item):
            print(('\t {:>%s} {}' % len(str(length)))
                  .format(i, item.key))

        for blob in d[word]:
            print(blob.id, blob.content_type, blob.key)
            progress = ''
            for i, item in enumerate(s):
                if i % 10000 == 0:
                    new_progress = '{:>4.1f}% ...'.format(100*i/length)
                    if new_progress != progress:
                        print(new_progress)
                        progress = new_progress

                if item.id == blob.id:
                    aliases.append((i, item))
                    print_item(i, item)
            print('100%')
            header = '{} {}'.format(blob.id, blob.content_type)
            print('='*len(header), '\n')
            print(header)
            print('-'*len(header))
            for i, item in aliases:
                print_item(i, item)


def _cli_get(args):
    with open(args.path) as s:
        _content_type, content = s.get(args.blob_id)
        sys.stdout.buffer.write(content)


def _p(i, *args, step=100, steps_per_line=50, fmt='{}'):
    line = steps_per_line*step
    if i % step == 0 and i != 0:
        sys.stdout.write('.')
        if i and i % line == 0:
            sys.stdout.write(fmt.format(*args))
        sys.stdout.flush()


def _cli_convert(args):
    import sys
    import time
    t0 = time.time()

    output_name = args.output

    output_dir = os.path.dirname(output_name)
    output_base = os.path.basename(output_name)
    output_base_noext, output_base_ext = os.path.splitext(output_base)

    DOT_SLOB = '.slob'

    if output_base_ext != DOT_SLOB:
        output_base = output_base + DOT_SLOB
        output_base_noext, output_base_ext = os.path.splitext(output_base)

    if args.split:
        output_dir = os.path.join(output_dir, output_base)
        if os.path.exists(output_dir):
            raise SystemExit('%r already exists' % output_dir)
        os.mkdir(output_dir)
    else:
        output_name = os.path.join(output_dir, output_base)
        if os.path.exists(output_name):
            raise SystemExit('%r already exists' % output_name)

    with open(args.path) as s:
        workdir = args.workdir
        encoding = args.encoding or s._header.encoding
        compression = (s._header.compression
                       if args.compression is None
                       else args.compression)
        min_bin_size = 1024*args.min_bin_size
        split = 1024*1024*args.split

        print('Mapping blobs to keys...')
        blob_to_refs = [collections.defaultdict(lambda: array.array('L'))
                        for i in range(len(s._store))]
        key_count = 0
        pp = functools.partial(_p, step=10000, fmt=' {:.2f}%/{:.2f}s\n')
        total_keys = len(s)
        blob_count = s.blob_count
        for i, ref in enumerate(s._refs):
            blob_to_refs[ref.bin_index][ref.item_index].append(i)
            key_count += 1
            pp(i, 100*key_count/total_keys, time.time() - t0)

        print('\nFound {} keys for {} blobs in {:.2f}s'
              .format(key_count, blob_count, time.time() - t0))

        print('Converting...')
        Mb = 1024*1024
        pp = functools.partial(_p, step=100, fmt=' {:.2f}%/{:.2f}Mb/{:.2f}s\n')

        def mkout(output):
            w = create(output,
                       workdir=workdir,
                       encoding=encoding,
                       compression=compression,
                       min_bin_size=min_bin_size)
            for name, value in s.tags.items():
                if not name in w.tags:
                    w.tag(name, value)
            size_header_and_tags = w.size_header() + w.size_tags()
            return w, size_header_and_tags

        def fin(t1, w, current_output):
            print('\nDone adding content to {1} in {0:.2f}s'
                  .format(time.time() - t1, current_output))
            print('Finalizing...')
            w.finalize()
            return time.time(), None, 0

        t1, w, size_header_and_tags = time.time(), None, 0
        current_size = 0
        volume_count = 0
        current_output = output_name
        current = 0
        for j, store_item in enumerate(s._store):
            for i in range(len(store_item.content_type_ids)):
                bin_index = j
                item_index = i
                current += 1
                content_type, content = s._store.get(bin_index, item_index)
                pp(current, 100*(current/blob_count), current_size/Mb, time.time() - t1)
                keys = []
                for k in blob_to_refs[bin_index][item_index]:
                    ref = s._refs[k]
                    keys.append((ref.key, ref.fragment))
                if w is None:
                    volume_count += 1
                    if split:
                        current_output = os.path.join(
                            output_dir,
                            ''.join((output_base_noext,
                                     '-{:02d}'.format(volume_count),
                                     output_base_ext)))
                    w, size_header_and_tags = mkout(current_output)
                w.add(content, *keys, content_type=content_type)
                current_size = (size_header_and_tags +
                                w.size_content_types() +
                                w.size_data())
                if split and current_size + min_bin_size >= split:
                    t1, w, size_header_and_tags = fin(t1, w, current_output)
        if w:
            fin(t1, w, current_output)

    print('\nDone in {0:.2f}s'.format(time.time() - t0))


def _arg_parser():
    parser = argparse.ArgumentParser()
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument('path', help='Slob path')

    parents = [parent]

    subparsers = parser.add_subparsers(help='sub-command')

    parser_find = subparsers.add_parser('find',
                                        parents=parents,
                                        help='Find keys', )
    parser_find.add_argument('query', help='Word to look up')
    parser_find.add_argument('--whole', action='store_true',
                             help='Match whole words only')
    parser_find.add_argument('-l', '--limit', type=int, default=10,
                             help='Maximum number of results to display')
    parser_find.set_defaults(func=_cli_find)

    parser_aliases = subparsers.add_parser('aliases',
                                            parents=parents,
                                            help='Find blob aliases', )
    parser_aliases.add_argument('query', help='Word to look up')
    parser_aliases.set_defaults(func=_cli_aliases)

    parser_get = subparsers.add_parser('get',
                                        parents=parents,
                                        help='Retrieve blob content', )
    parser_get.add_argument('blob_id',
                            type=int,
                            help='Id of blob to retrive (from output of "find")')

    parser_get.set_defaults(func=_cli_get)

    parser_info = subparsers.add_parser(
        'info',
        parents=parents,
        help='Inspect slob and print basic information about it')
    parser_info.set_defaults(func=_cli_info)

    parser_tag = subparsers.add_parser(
        'tag',
        help='List tags, view or edit tag value')
    parser_tag.add_argument('-n', '--name', default='',
                            help='Name of tag to view or edit')
    parser_tag.add_argument('-v', '--value',
                            help='Tag value to set')
    parser_tag.add_argument('filename',
                            help='Slob file name (split files are not supported)')
    parser_tag.set_defaults(func=_cli_tag)

    parser_convert = subparsers.add_parser(
        'convert',
        parents=parents,
        help=('Create new slob with the same content '
              'but different encoding and compression parameters '
              'or split into multiple slobs'))
    parser_convert.add_argument('output',
                                help='Name of slob file to create')

    parser_convert.add_argument('--workdir',
                                help='Directory where temporary files for '
                                'conversion should be created')

    parser_convert.add_argument('-e', '--encoding',
                                help=('Text encoding for the output slob. '
                                      'Default: same as source'))

    parser_convert.add_argument('-c', '--compression',
                                help=('Compression algorithm to use fot the output slob. '
                                      'Default: same as source'))

    parser_convert.add_argument('-b', '--min_bin_size', type=int,
                                help=('Minimum size of storage bin to compress in kilobytes. '
                                      'Default: %(default)s'),
                                default=384)

    parser_convert.add_argument('-s', '--split', type=int,
                                help=('Split output into multiple slob files no larger '
                                      'than specified number of megabytes. '
                                      'Default: %(default)s (do not split)'),
                                default=0)

    parser_convert.set_defaults(func=_cli_convert)

    return parser


def add_dir(slb, topdir, prefix='', include_only=None,
            mime_types=MIME_TYPES, print=print):
    print('Adding', topdir)
    for item in os.walk(topdir):
        dirpath, _dirnames, filenames = item
        for filename in filenames:
            full_path = os.path.join(dirpath, filename)
            rel_path = os.path.relpath(full_path, topdir)
            if include_only and not any(
                    rel_path.startswith(x) for x in include_only):
                print ('SKIPPING (not included): {!a}'.format(rel_path))
                continue
            _, ext = os.path.splitext(filename)
            ext = ext.lstrip(os.path.extsep)
            content_type = mime_types.get(ext.lower())
            if not content_type:
                print('SKIPPING (unknown content type): {!a}'.format(rel_path))
            else:
                with fopen(full_path, 'rb') as f:
                    content = f.read()
                    key = prefix + rel_path
                    print ('ADDING: {!a}'.format(key))
                    try:
                        key.encode(slb.encoding)
                    except UnicodeEncodeError:
                        print ('Failed to add, broken unicode in key: {!a}'.format(key))
                    else:
                        slb.add(content, key, content_type=content_type)


import time
from datetime import timedelta

class SimpleTimingObserver(object):

    def __init__(self, p=None):

        if not p is None:
            self.p = p

        self.times = {}

    def p(self, text):
        sys.stdout.write(text)
        sys.stdout.flush()

    def begin(self, name):
        self.times[name] = time.time()

    def end(self, name):
        t0 = self.times.pop(name)
        dt = timedelta(seconds=int(time.time() - t0))
        return dt

    def __call__(self, e):
        p = self.p
        begin = self.begin
        end = self.end
        if e.name == 'begin_finalize':
            p('\nFinished adding content in %s' % end('content'))
            p('\nFinalizing...')
            begin('finalize')
        if e.name == 'end_finalize':
            p('\nFinalized in %s' % end('finalize'))
        elif e.name == 'begin_resolve_aliases':
            p('\nResolving aliases...')
            begin('aliases')
        elif e.name == 'end_resolve_aliases':
            p('\nResolved aliases in %s' % end('aliases'))
        elif e.name == 'begin_sort':
            p('\nSorting...')
            begin('sort')
        elif e.name == 'end_sort':
            p(' sorted in %s' % end('sort'))


def main():
    parser = _arg_parser()
    args = parser.parse_args()
    if (hasattr(args, 'func')):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from formats_common import *
from pyglossary.text_reader import TextGlossaryReader
from pyglossary.text_utils import escapeNTB, unescapeNTB, splitByBarUnescapeNTB

enable = True
format = "Tabfile"
description = "Tabfile (txt, dic)"
extensions = [".txt", ".tab", ".dic"]
optionsProp = {
	"encoding": EncodingOption(),
	"writeInfo": BoolOption(),
	"resources": BoolOption(),
}
depends = {}


class Reader(TextGlossaryReader):
	def isInfoWord(self, word: str) -> bool:
		return word.startswith("#")

	def fixInfoWord(self, word: str) -> str:
		return word.lstrip("#")

	def nextPair(self) -> Tuple[str, str]:
		if not self._file:
			raise StopIteration
		line = self._file.readline()
		if not line:
			raise StopIteration
		line = line.strip()  # This also removes tailing newline
		if not line:
			return
		###
		word, tab, defi = line.partition("\t")
		if not tab:
			log.error(
				"Warning: line starting with \"%s\" has no tab!" % line[:10]
			)
			return
		###
		if self._glos.getPref("enable_alts", True):
			word = splitByBarUnescapeNTB(word)
			if len(word) == 1:
				word = word[0]
		else:
			word = unescapeNTB(word, bar=True)
		###
		defi = unescapeNTB(defi)
		###
		return word, defi


def write(
	glos: GlossaryType,
	filename: str,
	encoding: str = "utf-8",
	writeInfo: bool = True,
	resources: bool = True,
) -> bool:
	return glos.writeTabfile(
		filename,
		encoding=encoding,
		writeInfo=writeInfo,
		resources=resources,
	)



# -*- coding: utf-8 -*-

from formats_common import *

enable = False
format = "Test"
description = "Test Format File(.test)"
extensions = [".test", ".tst"]

# key is option/argument name, value is instance of Option
optionsProp = {}

depends = {}


def read(glos: GlossaryType, filename: str) -> bool:
	log.info("reading from format %s using plugin" % format)
	count = 100
	# get number of entries from input file(depending on your format)
	for i in range(count):
		# here get word and definition from file(depending on your format)
		word = "word_%s" % i
		defi = "definition %s" % i
		glos.addEntry(word, defi)
	# here read info from file and set to Glossary object
	glos.setInfo("name", "Test")
	glos.setInfo("descriptin", "Test glossary craeted by a PyGlossary plugin")
	glos.setInfo("author", "Me")
	glos.setInfo("copyright", "GPL")
	return True  # reading input file was succesfull


def write(glos: GlossaryType, filename: str) -> bool:
	log.info("writing to format %s using plugin" % format)
	for entry in glos:
		word = entry.getWord()
		defi = entry.getDefi()
		# here write word and defi to the output file (depending on
		# your format)
	# here read info from Glossaey object
	name = glos.getInfo("name")
	descriptin = glos.getInfo("descriptin")
	author = glos.getInfo("author")
	copyright = glos.getInfo("copyright")
	# if an info key doesn't exist, getInfo returns empty string
	# now write info to the output file (depending on your output format)
	return True  # writing output file was succesfull



# -*- coding: utf-8 -*-
#
# Copyright (C) Saeed Rasooli <saeed.gnu@gmail.com>
# Copyright (C) 2007 Mehdi Bayazee <Bayazee@Gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/gpl.txt>.
# Also avalable in /usr/share/common-licenses/GPL on Debian systems
# or /usr/share/licenses/common/GPL3/license.txt on ArchLinux


# Gregorian calendar:
# http://en.wikipedia.org/wiki/Gregorian_calendar

from datetime import datetime

from typing import (
	Tuple,
	Optional,
)

name = "gregorian"
desc = "Gregorian"
origLang = "en"

monthName = (
	"January",
	"February",
	"March",
	"April",
	"May",
	"June",
	"July",
	"August",
	"September",
	"October",
	"November",
	"December",
)

monthNameAb = (
	"Jan",
	"Feb",
	"Mar",
	"Apr",
	"May",
	"Jun",
	"Jul",
	"Aug",
	"Sep",
	"Oct",
	"Nov",
	"Dec",
)


def getMonthName(m: int, y: Optional[int] = None) -> str:
	return monthName.__getitem__(m - 1)


def getMonthNameAb(m, y: Optional[int] = None) -> str:
	return monthNameAb.__getitem__(m - 1)


def getMonthsInYear(y: int) -> int:
	return 12


epoch = 1721426
minMonthLen = 29
maxMonthLen = 31
avgYearLen = 365.2425  # FIXME

options = ()


def save() -> None:
	pass


def isLeap(y: int) -> bool:
	return y % 4 == 0 and not (y % 100 == 0 and y % 400 != 0)


def to_jd(year: int, month: int, day: int) -> int:
	if 0 < year < 10000:  # > 1.5x faster
		return datetime(year, month, day).toordinal() + 1721425

	if month <= 2:
		tm = 0
	elif isLeap(year):
		tm = -1
	else:
		tm = -2

	return (
		epoch - 1
		+ 365 * (year - 1)
		+ (year - 1) // 4
		- (year - 1) // 100
		+ (year - 1) // 400
		+ (367 * month - 362) // 12
		+ tm
		+ day
	)


def jd_to(jd: int) -> Tuple[int, int, int]:
	ordinal = int(jd) - 1721425
	if 0 < ordinal < 3652060:  # > 4x faster
		# datetime(9999, 12, 31).toordinal() == 3652059
		dt = datetime.fromordinal(ordinal)
		return (dt.year, dt.month, dt.day)

	# wjd = floor(jd - 0.5) + 0.5
	qc, dqc = divmod(jd - epoch, 146097)  # qc ~~ quadricent
	cent, dcent = divmod(dqc, 36524)
	quad, dquad = divmod(dcent, 1461)
	yindex = dquad // 365  # divmod(dquad, 365)[0]
	year = (
		qc * 400
		+ cent * 100
		+ quad * 4
		+ yindex
		+ (cent != 4 and yindex != 4)
	)
	yearday = jd - to_jd(year, 1, 1)

	if jd < to_jd(year, 3, 1):
		leapadj = 0
	elif isLeap(year):
		leapadj = 1
	else:
		leapadj = 2

	month = ((yearday + leapadj) * 12 + 373) // 367
	day = jd - to_jd(year, month, 1) + 1
	return int(year), int(month), int(day)


def getMonthLen(y: int, m: int) -> int:
	if m == 12:
		return to_jd(y + 1, 1, 1) - to_jd(y, 12, 1)
	else:
		return to_jd(y, m + 1, 1) - to_jd(y, m, 1)


J0001 = 1721426 # to_jd(1, 1, 1)
J1970 = 2440588 # to_jd(1970, 1, 1)

J0001_epoch = -62135621220



# -*- coding: utf-8 -*-

from typing import (
	Dict,
	Tuple,
	List,
	Optional,
	Iterator,
	Callable,
)

from collections import OrderedDict as odict

from .entry_base import BaseEntry
from .entry import Entry, DataEntry

class GlossaryType(object):
	def addEntryObj(self, entry: Entry) -> None:
		raise NotImplementedError

	def newEntry(self, word: str, defi: str, defiFormat: str = "") -> Entry:
		raise NotImplementedError

	def addEntry(self, word: str, defi: str, defiFormat: str = "") -> None:
		raise NotImplementedError

	def iterEntryBuckets(self, size: int) -> Iterator[BaseEntry]:
		raise NotImplementedError

	def setDefaultDefiFormat(self, defiFormat: str) -> None:
		raise NotImplementedError

	def getMostUsedDefiFormats(self, count: int = None) -> List[Tuple[str, int]]:
		raise NotImplementedError

	def iterInfo(self) -> Iterator[Tuple[str, str]]:
		raise NotImplementedError

	def getInfo(self, key: str) -> str:
		raise NotImplementedError

	def setInfo(self, key: str, value: str) -> None:
		raise NotImplementedError

	def getExtraInfos(self, excludeKeys: List[str]) -> odict:
		raise NotImplementedError

	def getPref(self, name: str, default: Optional[str]) -> Optional[str]:
		raise NotImplementedError

	def newDataEntry(self, fname: str, data: bytes) -> DataEntry:
		raise NotImplementedError

	def writeTxt(
		self,
		sep1: str,
		sep2: str,
		filename: str = "",
		writeInfo: bool = True,
		rplList: Optional[List[Tuple[str, str]]] = None,
		ext: str = ".txt",
		head: str = "",
		iterEntries: Optional[Iterator[BaseEntry]] = None,
		entryFilterFunc: Optional[Callable[[BaseEntry], Optional[BaseEntry]]] = None,
		outInfoKeysAliasDict: Optional[Dict[str, str]] = None,
		encoding: str = "utf-8",
		newline: str = "\n",
		resources: bool = True,
	) -> bool:
		raise NotImplementedError

	def writeTabfile(self, filename: str = "", **kwargs) -> None:
		raise NotImplementedError

	def iterSqlLines(
		self,
		filename: str = "",
		infoKeys: Optional[List] = None,
		addExtraInfo: bool = True,
		newline: str = "\\n",
		transaction: bool = False,
	) -> Iterator[str]:
		raise NotImplementedError


